from dataclasses import dataclass

from apischema import deserializer, serializer


@dataclass(frozen=True)
class ByteFormat:
    """An immutable dataclass for custom (de)serialization byte format strings."""

    format: bytes

    @serializer
    def serialize(self) -> str:
        """An apischema serialiazation method which returns a utf-8 decoded string.

        Returns:
            str: A utf-8 decoded string of the format.
        """
        return self.format.decode("utf-8")

    @deserializer
    @staticmethod
    def deserialize(data: str) -> "ByteFormat":
        """An apischema deserialization method builds from a utf-8 encoded string.

        Returns:
            ByteFormat: The deserialized ByteFormat.
        """
        return ByteFormat(data.encode("utf-8"))



import asyncio
import logging
from typing import Any

import aiozmq
import zmq

from tickit.core.adapter import Adapter, RaiseInterrupt
from tickit.core.device import Device

LOGGER = logging.getLogger(__name__)


class ZeroMQAdapter(Adapter):
    """An adapter for a ZeroMQ data stream."""

    zmq_host: str
    zmq_port: int
    running: bool

    _router: aiozmq.ZmqStream
    _dealer: aiozmq.ZmqStream

    def __init__(
        self, zmq_host: str = "127.0.0.1", zmq_port: int = 5555, running: bool = False
    ) -> None:
        """Initialize with default values."""
        super().__init__()
        self.zmq_host = zmq_host
        self.zmq_port = zmq_port
        self.running = running

    async def start_stream(self) -> None:
        """Start the ZeroMQ stream."""
        LOGGER.debug("Starting stream...")
        self._router = await aiozmq.create_zmq_stream(
            zmq.ROUTER, bind=f"tcp://{self.zmq_host}:{self.zmq_port}"
        )

        addr = list(self._router.transport.bindings())[0]
        self._dealer = await aiozmq.create_zmq_stream(zmq.DEALER, connect=addr)

        self._router.transport.setsockopt(zmq.LINGER, 0)
        self._dealer.transport.setsockopt(zmq.LINGER, 0)

    async def close_stream(self) -> None:
        """Close the ZeroMQ stream."""
        self._dealer.close()
        self._router.close()

        self.running = False

    def send_message(self, message: Any) -> None:
        """Send a message down the ZeroMQ stream.

        Sets up an asyncio task to put the message on the message queue, before
        being processed.

        Args:
            message (str): The message to send down the ZeroMQ stream.
        """
        asyncio.create_task(self._message_queue.put(message))

    async def run_forever(
        self, device: Device, raise_interrupt: RaiseInterrupt
    ) -> None:
        """Runs the ZeroMQ adapter continuously."""
        await super().run_forever(device, raise_interrupt)
        self._message_queue: asyncio.Queue = asyncio.Queue()
        await self.start_stream()
        self.running = True
        await self._process_message_queue()

    def check_if_running(self):
        """Returns the running state of the adapter."""
        return self.running

    async def _process_message_queue(self) -> None:
        running = True
        while running:
            message = await self._message_queue.get()
            await self._process_message(message)
            running = self.check_if_running()

    async def _process_message(self, message: str) -> None:
        if message is not None:
            LOGGER.debug("Data from ZMQ stream: {!r}".format(message))

            msg = (b"Data", str(message).encode("utf-8"))
            self._dealer.write(msg)
            data = await self._router.read()
            self._router.write(data)
            answer = await self._dealer.read()
            LOGGER.debug("Received {!r}".format(answer))
        else:
            LOGGER.debug("No message")



from dataclasses import dataclass

from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation

from .cryostream import CryostreamAdapter, CryostreamDevice


@dataclass
class Cryostream(ComponentConfig):
    """Cryostream simulation with TCP server."""

    host: str = "localhost"
    port: int = 25565

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=CryostreamDevice(),
            adapters=[CryostreamAdapter(host=self.host, port=self.port)],
        )



from abc import abstractmethod
from inspect import getmembers
from typing import AnyStr, AsyncIterable, Optional, Sequence, Tuple, get_type_hints

from tickit.adapters.interpreters.utils import wrap_as_async_iterable
from tickit.core.adapter import Adapter, Interpreter
from tickit.utils.compat.typing_compat import Protocol, runtime_checkable


@runtime_checkable
class Command(Protocol):
    """An interface for interperable commands."""

    #: A flag which indicates whether calling of the method should trigger an interrupt
    interrupt: bool

    @abstractmethod
    def parse(self, data: bytes) -> Optional[Sequence[AnyStr]]:
        """An abstract method which parses a message and extracts arguments.

        An abstract method which parses a message and produces arguments if a match is
        found, otherwise None is returned.

        Args:
            data (bytes): The message to be parsed.

        Returns:
            Optional[Sequence[AnyStr]]:
                A sequence of arguments extracted from the message if matched,
                otherwise None.
        """
        pass


class CommandInterpreter(Interpreter[AnyStr]):
    """An interpreter which routes to commands registered to adapter methods.

    An interpreter which attempts to parse messages according to the parse method of
    commands registered against adapter methods, if a match is found the method is
    called with the parsed arguments.
    """

    @staticmethod
    async def unknown_command() -> AsyncIterable[bytes]:
        """An asynchronous iterable of containing a single unknown command reply.

        Returns:
            AsyncIterable[bytes]:
                An asynchronous iterable of containing a single unknown command reply:
                "Request does not match any known command".
        """
        yield b"Request does not match any known command"

    async def handle(
        self, adapter: Adapter, message: AnyStr
    ) -> Tuple[AsyncIterable[AnyStr], bool]:
        """Matches the message to an adapter command and calls the corresponding method.

        An asynchronous method which handles a message by attempting to match the
        message against each of the registered commands, if a match is found the
        corresponding command is called and its reply is returned with an asynchronous
        iterable wrapper if required. If no match is found the unknown command message
        is returned with no request for interrupt.

        Args:
            adapter (Adapter): The adapter in which the function should be executed
            message (bytes): The message to be handled.

        Returns:
            Tuple[AsyncIterable[Union[str, bytes]], bool]:
                A tuple of the asynchronous iterable of reply messages and a flag
                indicating whether an interrupt should be raised by the adapter.
        """
        for _, method in getmembers(adapter):
            command = getattr(method, "__command__", None)
            if command is None:
                continue
            args = command.parse(message)
            if args is None:
                continue
            args = (
                argtype(arg)
                for arg, argtype in zip(args, get_type_hints(method).values())
            )
            resp = await method(*args)
            if not isinstance(resp, AsyncIterable):
                resp = wrap_as_async_iterable(resp)
            return resp, command.interrupt
        resp = CommandInterpreter.unknown_command()
        return resp, False



import logging
from typing import Any

from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.compat.typing_compat import TypedDict

LOGGER = logging.getLogger(__name__)


class SinkDevice(Device):
    """A simple device which can take any input and produces no output."""

    #: A typed mapping containing the 'input' input value
    Inputs: TypedDict = TypedDict("Inputs", {"input": Any})
    #: An empty typed mapping of device outputs
    Outputs: TypedDict = TypedDict("Outputs", {})

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """The update method which logs the inputs at debug level and produces no
        outputs.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which never contains any changes, and never
                requests a callback.
        """
        LOGGER.debug("Sunk {}".format({k: v for k, v in inputs.items()}))
        return DeviceUpdate(SinkDevice.Outputs(), None)


class Sink(ComponentConfig):
    """Arbitrary value sink that logs the value."""

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=SinkDevice(),
        )



import asyncio
import logging
import struct
from dataclasses import dataclass
from typing import AsyncIterable, Optional

from tickit.adapters.composed import ComposedAdapter
from tickit.adapters.interpreters.command import CommandInterpreter, RegexCommand
from tickit.adapters.servers.tcp import TcpServer
from tickit.core.adapter import Server
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.byte_format import ByteFormat
from tickit.utils.compat.typing_compat import TypedDict

LOGGER = logging.getLogger(__name__)


class RemoteControlledDevice(Device):
    """A trivial toy device which is controlled by an adapter."""

    #: An empty typed mapping of device inputs
    Inputs: TypedDict = TypedDict("Inputs", {})
    #: A typed mapping containing the 'observed' output value
    Outputs: TypedDict = TypedDict("Outputs", {"observed": float})

    def __init__(
        self,
        initial_observed: float = 0,
        initial_unobserved: float = 42,
        initial_hidden: float = 3.14,
    ) -> None:
        """A RemoteControlled constructor which configures various initial values.

        Args:
            initial_observed (float): The initial value of the observed device property.
                Defaults to 0.
            initial_unobserved (float): The initial value of the unobserved device
                property. Defaults to 42.
            initial_hidden (float): The initial value of the hidden device property.
                Defaults to 3.14.
        """
        self.observed = initial_observed
        self.unobserved = initial_unobserved
        self.hidden = initial_hidden

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """The update method which produces an output mapping containing the observed
        value.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the observed value, the device
                never requests a callback.
        """
        return DeviceUpdate(self.Outputs(observed=self.observed), None)


class RemoteControlledAdapter(ComposedAdapter[bytes]):
    """A trivial composed adapter which gets and sets device properties."""

    device: RemoteControlledDevice

    def __init__(
        self,
        server: Server,
    ) -> None:
        """A constructor of the Shutter adapter, which builds the configured server.

        Args:
            device (Device): The device which this adapter is attached to.
            raise_interrupt (Callable): A callback to request that the device is
                updated immediately.
            server (Server): The immutable data container used to configure a
                server.
        """
        super().__init__(
            server,
            CommandInterpreter(),
        )

    async def on_connect(self) -> AsyncIterable[Optional[bytes]]:
        """An on_connect method which continiously sends the unobserved value to the
        client.

        Returns:
            AsyncIterable[bytes]:
                An asynchronous iterable which regularly outputs the unobserved value.
        """
        while True:
            await asyncio.sleep(5.0)
            yield "U is {}".format(self.device.unobserved).encode("utf-8")

    @RegexCommand(b"\x01")
    async def get_observed_bytes(self) -> bytes:
        """A regex bytes command which returns the byte encoded value of observed.

        Returns:
            bytes: The big endian float encoded value of observed.
        """
        return struct.pack(">f", self.device.observed)

    @RegexCommand(r"O", format="utf-8")
    async def get_observed_str(self) -> bytes:
        """A regex string command which returns the utf-8 encoded value of observed.

        Returns:
            bytes: The utf-8 encoded value of observed.
        """
        return str(self.device.observed).encode("utf-8")

    @RegexCommand(b"\x01(.{4})", interrupt=True)
    async def set_observed_bytes(self, value: bytes) -> bytes:
        """A regex bytes command which sets and echos the value of observed.

        Args:
            value (bytes): The new big endian float encoded value of observed.

        Returns:
            bytes: The big endian float encoded value of observed.
        """
        self.device.observed = struct.unpack(">f", value)[0]
        return struct.pack(">f", self.device.observed)

    @RegexCommand(r"O=(\d+\.?\d*)", interrupt=True, format="utf-8")
    async def set_observed_str(self, value: float) -> bytes:
        """A regex string command which sets and echos the value of observed.

        Args:
            value (int): The new value of observed.

        Returns:
            bytes: The utf-8 encoded value of observed.
        """
        self.device.observed = value
        return "Observed set to {}".format(self.device.observed).encode("utf-8")

    @RegexCommand(b"\x02")
    async def get_unobserved_bytes(self) -> bytes:
        """A regex bytes command which returns the byte encoded value of unobserved.

        Returns:
            bytes: The big endian float encoded value of unobserved.
        """
        return struct.pack(">f", self.device.unobserved)

    @RegexCommand(r"U", format="utf-8")
    async def get_unobserved_str(self) -> bytes:
        """A regex string command which returns the utf-8 encoded value of unobserved.

        Returns:
            bytes: The utf-8 encoded value of unobserved.
        """
        return str(self.device.unobserved).encode("utf-8")

    @RegexCommand(b"\x02(.{4})")
    async def set_unobserved_bytes(self, value: bytes) -> bytes:
        """A regex bytes command which sets and echos the value of unobserved.

        Args:
            value (bytes): The new big endian float encoded value of unobserved.

        Returns:
            bytes: The big endian float encoded value of unobserved.
        """
        self.device.unobserved = struct.unpack(">f", value)[0]
        return struct.pack(">f", self.device.unobserved)

    @RegexCommand(r"U=(\d+\.?\d*)", format="utf-8")
    async def set_unobserved_str(self, value: float) -> bytes:
        """A regex string command which sets and echos the value of unobserved.

        Args:
            value (int): The new value of unobserved.

        Returns:
            bytes: The utf-8 encoded value of unobserved.
        """
        self.device.unobserved = value
        return "Unobserved set to {}".format(self.device.unobserved).encode("utf-8")

    @RegexCommand(chr(0x1F95A), format="utf-8")
    async def misc(self) -> bytes:
        """A regex string command which returns a utf-8 encoded character.

        Returns:
            bytes: A utf-8 encoded character.
        """
        return chr(0x1F430).encode("utf-8")

    @RegexCommand(r"H=(\d+\.?\d*)", format="utf-8")
    async def set_hidden(self, value: float) -> None:
        """A regex string command which sets the value of hidden.

        Args:
            value (float): The new value of hidden.
        """
        LOGGER.info("Hidden set to {}".format(self.device.hidden))

    @RegexCommand(r"H", format="utf-8")
    async def get_hidden(self) -> None:
        """A regex string command which returns nothing, hidden cannot be shown."""
        pass

    @RegexCommand(r"O\?(\d+)", format="utf-8")
    async def yield_observed(self, n: int = 10) -> AsyncIterable[bytes]:
        """A regex string command which sends observed numerous times.

        Args:
            n (int): The number of times which observed should be sent. Defaults to 10.

        Returns:
            AsyncIterable[bytes]:
                An asynchronous iterable which yields the value of observed the
                specified number of times with a fixed delay.
        """
        for i in range(1, int(n)):
            await asyncio.sleep(1.0)
            yield "Observed is {}".format(self.device.observed).encode("utf-8")


@dataclass
class RemoteControlled(ComponentConfig):
    """Thing you can poke over TCP."""

    format: ByteFormat = ByteFormat(b"%b\r\n")

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=RemoteControlledDevice(),
            adapters=[RemoteControlledAdapter(TcpServer())],
        )



import logging
from dataclasses import fields

from apischema import serialize
from typing_extensions import TypedDict

from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.devices.eiger.data.dummy_image import Image
from tickit.devices.eiger.eiger_schema import AccessMode, Value
from tickit.devices.eiger.eiger_settings import EigerSettings
from tickit.devices.eiger.filewriter.filewriter_config import FileWriterConfig
from tickit.devices.eiger.filewriter.filewriter_status import FileWriterStatus
from tickit.devices.eiger.monitor.monitor_config import MonitorConfig
from tickit.devices.eiger.monitor.monitor_status import MonitorStatus
from tickit.devices.eiger.stream.stream_config import StreamConfig
from tickit.devices.eiger.stream.stream_status import StreamStatus

from .eiger_status import EigerStatus, State

LOGGER = logging.getLogger(__name__)


class EigerDevice(Device):
    """A device class for the Eiger detector."""

    settings: EigerSettings
    status: EigerStatus

    #: An empty typed mapping of input values
    Inputs: TypedDict = TypedDict("Inputs", {"flux": float})
    #: A typed mapping containing the 'value' output value
    Outputs: TypedDict = TypedDict("Outputs", {})

    def __init__(
        self,
    ) -> None:
        """An Eiger device constructor.

        An Eiger device constructor which configures the default settings and various
        states of the device.
        """
        self.settings = EigerSettings()
        self.status = EigerStatus()

        self.stream_status = StreamStatus()
        self.stream_config = StreamConfig()
        self.stream_callback_period = SimTime(int(1e9))

        self.filewriter_status: FileWriterStatus = FileWriterStatus()
        self.filewriter_config: FileWriterConfig = FileWriterConfig()
        self.filewriter_callback_period = SimTime(int(1e9))

        self.monitor_status: MonitorStatus = MonitorStatus()
        self.monitor_config: MonitorConfig = MonitorConfig()
        self.monitor_callback_period = SimTime(int(1e9))

    async def initialize(self) -> None:
        """Function to initialise the Eiger."""
        self._set_state(State.IDLE)

    async def arm(self) -> None:
        """Function to arm the Eiger."""
        self._set_state(State.READY)

        header_detail = self.stream_config["header_detail"]["value"]

        json = {
            "htype": "dheader-1.0",
            "series": "<id>",
            "header_detail": header_detail,
        }
        if header_detail != "none":
            config_json = {}
            disallowed_configs = ["flatfield", "pixelmask" "countrate_correction_table"]
            for field_ in fields(self.settings):
                if field_.name not in disallowed_configs:
                    config_json[field_.name] = vars(self.settings)[field_.name]

        LOGGER.debug(json)
        LOGGER.debug(config_json)

    async def disarm(self) -> None:
        """Function to disarm the Eiger."""
        self._set_state(State.IDLE)

        json = {"htype": "dseries_end-1.0", "series": "<id>"}

        LOGGER.debug(json)

    async def trigger(self) -> str:
        """Function to trigger the Eiger.

        If the detector is in an external trigger mode, this is disabled as
        this software command interface only works for internal triggers.
        """
        trigger_mode = self.settings.trigger_mode
        state = self.status.state

        if state == State.READY and trigger_mode == "ints":
            self._set_state(State.ACQUIRE)

            for idx in range(0, self.settings.nimages):
                aquired = Image.create_dummy_image(idx)

                header_json = {
                    "htype": "dimage-1.0",
                    "series": "<series id>",
                    "frame": aquired.index,
                    "hash": aquired.hash,
                }

                json2 = {
                    "htype": "dimage_d-1.0",
                    "shape": "[x,y,(z)]",
                    "type": aquired.dtype,
                    "encoding": aquired.encoding,
                    "size": len(aquired.data),
                }

                json3 = {
                    "htype": "dconfig-1.0",
                    "start_time": "<start_time>",
                    "stop_time": "<stop_time>",
                    "real_time": "<real_time>",
                }

                LOGGER.debug(header_json)
                LOGGER.debug(json2)
                LOGGER.debug(json3)

            return "Aquiring Data from Eiger..."
        else:
            return (
                f"Ignoring trigger, state={self.status.state},"
                f"trigger_mode={trigger_mode}"
            )

    async def cancel(self) -> None:
        """Function to stop the data acquisition.

        Function to stop the data acquisition, but only after the next
        image is finished.
        """
        self._set_state(State.READY)

        header_json = {"htype": "dseries_end-1.0", "series": "<id>"}

        LOGGER.debug(header_json)

    async def abort(self) -> None:
        """Function to abort the current task on the Eiger."""
        self._set_state(State.IDLE)

        header_json = {"htype": "dseries_end-1.0", "series": "<id>"}

        LOGGER.debug(header_json)

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """Generic update function to update the values of the ExampleHTTPDevice.

        Args:
            time (SimTime): The simulation time in nanoseconds.
            inputs (Inputs): A TypedDict of the inputs to the ExampleHTTPDevice.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the value of the device
                variables.
        """
        current_flux = inputs["flux"]

        intensity_scale = (current_flux / 100) * 100
        LOGGER.debug(f"Relative beam intensity: {intensity_scale}")

        return DeviceUpdate(self.Outputs(), None)

    def get_state(self):  # TODO: Add return type hint
        """Returns the current state of the Eiger.

        Returns:
            State: The state of the Eiger.
        """
        val = self.status.state
        allowed = [s.value for s in State]
        return serialize(
            Value(
                val,
                AccessMode.STRING,  # type: ignore
                access_mode=AccessMode.READ_ONLY,
                allowed_values=allowed,
            )
        )

    def _set_state(self, state: State):
        self.status.state = state



from dataclasses import dataclass

from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation

from .current import CurrentDevice
from .femto import FemtoAdapter, FemtoDevice


@dataclass
class Femto(ComponentConfig):
    """Femto simulation with EPICS IOC."""

    initial_gain: float = 2.5
    initial_current: float = 0.0
    db_file: str = "tickit/devices/femto/record.db"
    ioc_name: str = "FEMTO"

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=FemtoDevice(
                initial_gain=self.initial_gain, initial_current=self.initial_current
            ),
            adapters=[FemtoAdapter(db_file=self.db_file, ioc_name=self.ioc_name)],
        )


@dataclass
class Current(ComponentConfig):
    """Simulated current source."""

    callback_period: int = int(1e9)

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=CurrentDevice(callback_period=self.callback_period),
        )



import os
import re
from abc import abstractmethod
from dataclasses import dataclass
from tempfile import NamedTemporaryFile
from typing import Any, Callable, Dict

from softioc import builder, softioc

from tickit.core.adapter import Adapter, RaiseInterrupt
from tickit.core.device import Device

from .ioc_manager import notify_adapter_ready, register_adapter


@dataclass(frozen=True)
class InputRecord:
    """A data container representing an EPICS input record."""

    name: str
    set: Callable
    get: Callable


@dataclass
class OutputRecord:
    """A data container representing an EPICS output record."""

    name: str


class EpicsAdapter(Adapter):
    """An adapter implementation which acts as an EPICS IOC."""

    def __init__(self, db_file: str, ioc_name: str) -> None:
        """An EpicsAdapter constructor which stores the db_file path and the IOC name.

        Args:
            db_file (str): The path to the db_file.
            ioc_name (str): The name of the EPICS IOC.
        """
        self.db_file = db_file
        self.ioc_name = ioc_name
        self.interrupt_records: Dict[InputRecord, Callable[[], Any]] = {}
        self.ioc_num = register_adapter()

    def link_input_on_interrupt(
        self, record: InputRecord, getter: Callable[[], Any]
    ) -> None:
        """Adds a record and a getter to the mapping of interrupting records.

        Args:
            record (InputRecord): The record to be added.
            getter (Callable[[], Any]): The getter handle.
        """
        self.interrupt_records[record] = getter

    def after_update(self) -> None:
        """Updates IOC records immediately following a device update."""
        for record, getter in self.interrupt_records.items():
            current_value = getter()
            record.set(current_value)
            print("Record {} updated to : {}".format(record.name, current_value))

    @abstractmethod
    def on_db_load(self) -> None:
        """Customises records that have been loaded in to suit the simulation."""
        raise NotImplementedError

    def load_records_without_DTYP_fields(self):
        """Loads the records without DTYP fields."""
        with open(self.db_file, "rb") as inp:
            with NamedTemporaryFile(suffix=".db", delete=False) as out:
                for line in inp.readlines():
                    if not re.match(rb"\s*field\s*\(\s*DTYP", line):
                        out.write(line)

        softioc.dbLoadDatabase(out.name, substitutions=f"device={self.ioc_name}")
        os.unlink(out.name)

    async def run_forever(
        self, device: Device, raise_interrupt: RaiseInterrupt
    ) -> None:
        """Runs the server continously."""
        await super().run_forever(device, raise_interrupt)
        builder.SetDeviceName(self.ioc_name)
        self.load_records_without_DTYP_fields()
        self.on_db_load()
        builder.UnsetDevice()
        notify_adapter_ready(self.ioc_num)



import asyncio
import logging

import click
from click.core import Context

from tickit.core.management.event_router import InverseWiring
from tickit.core.management.schedulers.master import MasterScheduler
from tickit.core.runner import run_all_forever
from tickit.core.state_interfaces.state_interface import get_interface, interfaces
from tickit.utils.configuration.loading import read_configs


@click.group(invoke_without_command=True)
@click.version_option()
@click.option(
    "--log-level",
    default="DEBUG",
    type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]),
    help="The minimum level of severity for a log message to be printed to the console",
)
@click.pass_context
def main(ctx: Context, log_level: str):
    """The command line argument root, allowing for mode and log level configuration.

    Args:
        ctx (Context): The click context.
        log_level (str): The minimum logging level to be displayed.
    """
    logging.basicConfig(level=log_level)

    if ctx.invoked_subcommand is None:
        click.echo(main.get_help(ctx))


@main.command(help="run a single simulated component")
@click.argument("component")
@click.argument("config_path")
@click.option("--backend", default="kafka", type=click.Choice(list(interfaces(True))))
def component(config_path: str, component: str, backend: str) -> None:
    """Runs a single simulated component from a configuration file.

    Args:
        config_path (str): The path to the configuration file.
        component (str): The name of the component to be run.
        backend (str): The message broker to be used.
    """
    configs = read_configs(config_path)
    config = [config for config in configs if config.name == component]
    assert len(config) == 1, f"Expected only one component {component}"
    asyncio.run(run_all_forever([config[0]().run_forever(*get_interface(backend))]))


@main.command(help="run the simulation scheduler")
@click.argument("config_path")
@click.option("--backend", default="kafka", type=click.Choice(list(interfaces(True))))
def scheduler(config_path: str, backend: str) -> None:
    """Runs the simulation master scheduler from a configuration file.

    Args:
        config_path (str): The path to the configuration file.
        backend (str): The message broker to be used.
    """
    configs = read_configs(config_path)
    inverse_wiring = InverseWiring.from_component_configs(configs)
    scheduler = MasterScheduler(inverse_wiring, *get_interface(backend))
    asyncio.run(run_all_forever([scheduler.run_forever()]))


@main.command(help="run a collection of devices with a scheduler")
@click.argument("config_path")
@click.option(
    "--backend", default="internal", type=click.Choice(list(interfaces(False)))
)
def all(config_path: str, backend: str) -> None:
    """Runs all components and the master scheduler from a configuration file.

    Args:
        config_path (str): The path to the configuration file.
        backend (str): The message broker to be used.
    """
    configs = read_configs(config_path)
    inverse_wiring = InverseWiring.from_component_configs(configs)
    scheduler = MasterScheduler(inverse_wiring, *get_interface(backend))
    asyncio.run(
        run_all_forever(
            [config().run_forever(*get_interface(backend)) for config in configs]
            + [scheduler.run_forever()]
        )
    )



import pytest
from mock import AsyncMock

from tickit.core.runner import run_all_forever


@pytest.mark.asyncio
async def test_run_forever_all_awaits():
    awaitable = AsyncMock()
    await run_all_forever([awaitable()])
    awaitable.assert_awaited_once_with()



from dataclasses import is_dataclass
from typing import Optional

from tickit.core.typedefs import Changes, ComponentID, Input, Output, SimTime


def test_input_is_dataclass():
    assert is_dataclass(Input)


def test_input_has_target():
    assert ComponentID == Input.__annotations__["target"]


def test_input_has_time():
    assert SimTime == Input.__annotations__["time"]


def test_input_has_changes():
    assert Changes == Input.__annotations__["changes"]


def test_output_is_dataclass():
    assert is_dataclass(Output)


def test_output_has_source():
    assert ComponentID == Output.__annotations__["source"]


def test_output_has_time():
    assert SimTime == Output.__annotations__["time"]


def test_output_has_changes():
    assert Changes == Output.__annotations__["changes"]


def test_output_has_call_in():
    assert Optional[SimTime] == Output.__annotations__["call_at"]



from typing import Dict, Iterable

import pytest
from immutables import Map
from mock import AsyncMock, Mock, create_autospec, patch

from tickit.core.management.event_router import InverseWiring, Wiring
from tickit.core.management.schedulers.slave import SlaveScheduler
from tickit.core.management.ticker import Ticker
from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer
from tickit.core.typedefs import (
    Changes,
    ComponentID,
    ComponentPort,
    Input,
    Output,
    PortID,
    SimTime,
)
from tickit.utils.topic_naming import input_topic


@pytest.fixture
def mock_wiring() -> Mock:
    mock = create_autospec(InverseWiring, instance=True)
    mock[ComponentID("expose")] = create_autospec({})
    return mock


@pytest.fixture
def mock_state_consumer() -> Mock:
    return create_autospec(StateConsumer, instance=False)


@pytest.fixture
def mock_state_producer() -> Mock:
    return create_autospec(StateProducer, instance=False)


@pytest.fixture
def mock_raise_interrupt() -> Mock:
    async def interrupt():
        pass

    return create_autospec(interrupt)


@pytest.fixture
def patch_ticker() -> Iterable[Mock]:
    with patch("tickit.core.management.schedulers.base.Ticker", autospec=True) as mock:
        mock.return_value = AsyncMock(spec=Ticker)
        mock.return_value.time = SimTime(41)
        yield mock


@pytest.fixture
def expose() -> Dict[PortID, ComponentPort]:
    return {PortID("42"): ComponentPort(ComponentID("test_component"), PortID("53"))}


@pytest.fixture
def slave_scheduler(
    mock_wiring, mock_state_consumer, mock_state_producer, expose, mock_raise_interrupt
) -> SlaveScheduler:
    return SlaveScheduler(
        mock_wiring,
        mock_state_consumer,
        mock_state_producer,
        expose,
        mock_raise_interrupt,
    )


def test_construct_slave_scheduler(slave_scheduler, expose):
    pass


@pytest.mark.parametrize(
    "wiring",
    [
        pytest.param(Wiring(), id="Wiring"),
        pytest.param(InverseWiring(), id="InverseWiring"),
    ],
)
def test_slave_scheduler_add_exposing_wiring_static_method(wiring):
    """Method to test the static 'add_exposing_wiring' method of 'SlaveScheduler'.

    This test passes if:
    1) The object returned by the method is an instance of 'InverseWiring' regardless
        of the type of the given 'wiring' object;
    2) The entry for 'ComponentID("expose")' in the 'InverseWiring' object is updated
        with the ports to expose.
    """
    expose = {PortID("42"): ComponentPort(ComponentID("test_component"), PortID("53"))}
    result = SlaveScheduler.add_exposing_wiring(wiring, expose)

    assert isinstance(result, InverseWiring)
    assert result[ComponentID("expose")] == expose


@pytest.fixture
def mock_ticker() -> Mock:
    return AsyncMock(spec_set=Ticker)


@pytest.mark.asyncio
async def test_slave_scheduler_update_external_component(
    slave_scheduler: SlaveScheduler,
    mock_ticker: Ticker,
):
    target = ComponentID("external")
    input = Input(
        target=target,
        time=SimTime(9),
        changes=Changes(Map({PortID("31"): 30})),
    )

    slave_scheduler.ticker = mock_ticker
    slave_scheduler.input_changes = input.changes
    await slave_scheduler.update_component(input)
    mock_ticker.propagate.assert_awaited_once_with(  # type: ignore
        Output(target, input.time, input.changes, None)
    )


@pytest.mark.asyncio
async def test_slave_scheduler_update_exposed_component(
    slave_scheduler: SlaveScheduler,
    mock_ticker: Ticker,
):
    target = ComponentID("expose")
    input = Input(
        target=target,
        time=SimTime(9),
        changes=Changes(Map({PortID("31"): 30})),
    )

    slave_scheduler.ticker = mock_ticker
    await slave_scheduler.update_component(input)
    mock_ticker.propagate.assert_awaited_once_with(  # type: ignore
        Output(target, input.time, Changes(Map()), None)
    )


@pytest.mark.asyncio
async def test_slave_scheduler_update_other_component(
    slave_scheduler: SlaveScheduler,
    mock_ticker: Ticker,
    mock_state_producer: AsyncMock,
):
    target = ComponentID("other")
    input = Input(
        target=target,
        time=SimTime(9),
        changes=Changes(Map({PortID("31"): 30})),
    )

    slave_scheduler.ticker = mock_ticker
    slave_scheduler.state_producer = mock_state_producer
    await slave_scheduler.update_component(input)
    mock_state_producer.produce.assert_awaited_once_with(input_topic(target), input)


@pytest.mark.asyncio
async def test_slave_scheduler_run_forever_method(slave_scheduler: SlaveScheduler):
    await slave_scheduler.run_forever()


@pytest.mark.asyncio
async def test_slave_scheduler_on_tick_method(
    slave_scheduler: SlaveScheduler, mock_ticker: Mock
):
    changes = Changes(Map({PortID("67"): 67}))
    slave_scheduler.ticker = mock_ticker
    output_changes, call_at = await slave_scheduler.on_tick(SimTime(8), changes)
    assert output_changes == Changes(Map())
    assert call_at is None


@pytest.mark.asyncio
async def test_slave_scheduler_on_tick_method_with_wakeups(
    slave_scheduler: SlaveScheduler, mock_ticker: Mock
):
    changes = Changes(Map({PortID("67"): 67}))
    slave_scheduler.ticker = mock_ticker
    slave_scheduler.wakeups = {
        ComponentID("first"): SimTime(1),
        ComponentID("second"): SimTime(2),
    }
    output_changes, call_at = await slave_scheduler.on_tick(SimTime(1), changes)
    assert output_changes == Changes(Map())
    assert call_at == SimTime(2)


@pytest.mark.asyncio
async def test_slave_scheduler_schedule_interrupt_method(
    slave_scheduler: SlaveScheduler,
):
    interrupt = ComponentID("interrupt")
    await slave_scheduler.schedule_interrupt(interrupt)
    assert interrupt in slave_scheduler.interrupts



from tickit.core.typedefs import ComponentID


def valid_component_id(component: ComponentID) -> None:
    """Checks the validity of a ComponentID for generating a topic name.

    Args:
        component (ComponentID): The ComponentID to check.

    Raises:
        ValueError: The component name is empty.
    """
    if not component:
        raise ValueError


def output_topic(component: ComponentID) -> str:
    """Creates the output topic name for a given component.

    Args:
        component (ComponentID): The component for which an output topic name is
            required.

    Returns:
        str: The output topic name of the component.
    """
    valid_component_id(component)
    return "tickit-" + component + "-out"


def input_topic(component: ComponentID) -> str:
    """Creates returns the input topic name for a given component.

    Args:
        component (ComponentID): The component for which an input topic name is
            required.

    Returns:
        str: The input topic name of the component.
    """
    valid_component_id(component)
    return "tickit-" + component + "-in"



import asyncio
from dataclasses import dataclass, field
from typing import Awaitable, Callable, Dict, Hashable, List, Mapping, Type, cast

from immutables import Map

from tickit.core.adapter import Adapter
from tickit.core.components.component import BaseComponent
from tickit.core.device import Device, DeviceUpdate
from tickit.core.runner import run_all
from tickit.core.state_interfaces import StateConsumer, StateProducer
from tickit.core.typedefs import Changes, ComponentID, SimTime, State

InterruptHandler = Callable[[], Awaitable[None]]


@dataclass
class DeviceSimulation(BaseComponent):
    """A component containing a device and the corresponding adapters.

    A component which thinly wraps a device and the corresponding adapters, this
    component delegates core behaviour to the update method of the device, whilst
    allowing adapters to raise interrupts.
    """

    name: ComponentID
    device: Device
    adapters: List[Adapter] = field(default_factory=list)
    last_outputs: State = field(init=False, default_factory=lambda: State({}))
    device_inputs: Dict[str, Hashable] = field(init=False, default_factory=dict)

    async def run_forever(
        self, state_consumer: Type[StateConsumer], state_producer: Type[StateProducer]
    ) -> None:
        """Sets up state interfaces, runs adapters and blocks until any complete."""
        tasks = run_all(
            adapter.run_forever(self.device, self.raise_interrupt)
            for adapter in self.adapters
        )
        await super().run_forever(state_consumer, state_producer)
        if tasks:
            await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

    async def on_tick(self, time: SimTime, changes: Changes) -> None:
        """Delegates core behaviour to the device and calls adapter on_update.

        An asynchronous method which updates device inputs according to external
        changes, delegates core behaviour to the device update method, informs
        Adapters of the update, computes changes to the state of the component
        and sends the resulting Output.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            changes (Changes): A mapping of changed component inputs and their new
                values.
        """
        self.device_inputs = {
            **self.device_inputs,
            **cast(Mapping[str, Hashable], changes),
        }
        device_update: DeviceUpdate = self.device.update(
            SimTime(time), self.device_inputs
        )
        for adapter in self.adapters:
            adapter.after_update()
        out_changes = Changes(
            Map(
                {
                    k: v
                    for k, v in device_update.outputs.items()
                    if k not in self.last_outputs or not self.last_outputs[k] == v
                }
            )
        )
        self.last_outputs = device_update.outputs
        await self.output(time, out_changes, device_update.call_at)



import asyncio
import itertools
import logging
from typing import Set

from softioc import asyncio_dispatcher, builder, softioc

LOGGER = logging.getLogger(__name__)

#: Name for the global Tickit IOC, will prefix some meta PVs.
_TICKIT_IOC_NAME: str = "TICKIT_IOC"

#: Ids of all adapters currently registered but not ready.
_REGISTERED_ADAPTER_IDS: Set[int] = set()

#: Iterator of unique IDs for new adapters
_ID_COUNTER: itertools.count = itertools.count()


def register_adapter() -> int:
    """Register a new adapter that may be creating records for the process-wide IOC.

    The IOC will not be initialized until all registered adapters have notified that
    they are ready.

    Returns:
        int: A unique ID for this adapter to use when notifiying that it is ready.
    """
    adapter_id = next(_ID_COUNTER)
    LOGGER.debug(f"New IOC adapter registering with ID: {adapter_id}")
    _REGISTERED_ADAPTER_IDS.add(adapter_id)
    return adapter_id


def notify_adapter_ready(adapter_id: int) -> None:
    """Notify the builder that a particular adpater has made all the records it needs.

    Once all registered adapters have notified, the IOC will start.

    Args:
        adapter_id (int): Unique ID of the adapter
    """
    _REGISTERED_ADAPTER_IDS.remove(adapter_id)
    LOGGER.debug(f"IOC adapter #{adapter_id} reports ready")
    if not _REGISTERED_ADAPTER_IDS:
        LOGGER.debug("All registered adapters are ready, starting IOC")
        _build_and_run_ioc()


def _build_and_run_ioc() -> None:
    """Build an EPICS python soft IOC for the adapter."""
    LOGGER.debug("Initializing database")

    # Records become immutable after this point
    builder.SetDeviceName(_TICKIT_IOC_NAME)
    softioc.devIocStats(_TICKIT_IOC_NAME)
    builder.LoadDatabase()

    LOGGER.debug("Starting IOC")
    event_loop = asyncio.get_event_loop()
    dispatcher = asyncio_dispatcher.AsyncioDispatcher(event_loop)
    softioc.iocInit(dispatcher)
    # dbl directly prints out all record names, so we have to check
    # the log level in order to only do it in DEBUG.
    if LOGGER.level <= logging.DEBUG:
        softioc.dbl()
    LOGGER.debug("IOC started")



import asyncio
from dataclasses import dataclass
from typing import Dict, List, Type

from tickit.core.components.component import BaseComponent, Component, ComponentConfig
from tickit.core.management.event_router import InverseWiring
from tickit.core.management.schedulers.slave import SlaveScheduler
from tickit.core.runner import run_all
from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer
from tickit.core.typedefs import Changes, ComponentID, ComponentPort, PortID, SimTime


@dataclass
class SystemSimulationComponent(BaseComponent):
    """A component containing a slave scheduler and several components.

    A component which acts as a nested tickit simulation by wrapping a slave scheduler
    and a set of internal components, this component delegates core behaviour to the
    components within it, whilst outputting their requests for wakeups and interrupts.
    """

    name: ComponentID

    #: A list of immutable component configuration data containers, used to
    #: construct internal components.
    components: List[ComponentConfig]
    #: A mapping of outputs which the system simulation exposes and the
    #: corresponding output of an internal component.
    expose: Dict[PortID, ComponentPort]

    async def run_forever(
        self, state_consumer: Type[StateConsumer], state_producer: Type[StateProducer]
    ) -> None:
        """Sets up state interfaces, the scheduler, and components and blocks until any
        complete.

        An asynchronous method starts the run_forever method of each component, runs
        the scheduler, and sets up externally facing state interfaces. The method
        blocks until and of the components or the scheduler complete.
        """
        inverse_wiring = InverseWiring.from_component_configs(self.components)
        self.scheduler = SlaveScheduler(
            inverse_wiring,
            state_consumer,
            state_producer,
            self.expose,
            self.raise_interrupt,
        )
        tasks = run_all(
            component().run_forever(state_consumer, state_producer)
            for component in self.components
        ) + run_all([self.scheduler.run_forever()])
        await super().run_forever(state_consumer, state_producer)
        await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)

    async def on_tick(self, time: SimTime, changes: Changes) -> None:
        """Delegates core behaviour to the slave scheduler.

        An asynchronous method which delegates core behaviour of computing changes and
        determining a callback period to the slave shceduler and sends the resulting
        Output.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            changes (Changes): A mapping of changed component inputs and their new
                values.
        """
        output_changes, call_in = await self.scheduler.on_tick(time, changes)
        await self.output(time, output_changes, call_in)


@dataclass
class SystemSimulation(ComponentConfig):
    """Simulation of a nested set of components."""

    name: ComponentID
    inputs: Dict[PortID, ComponentPort]
    components: List[ComponentConfig]
    expose: Dict[PortID, ComponentPort]

    def __call__(self) -> Component:  # noqa: D102
        return SystemSimulationComponent(
            name=self.name,
            components=self.components,
            expose=self.expose,
        )



import pytest

from tickit.devices.eiger.eiger_settings import EigerSettings, KA_Energy

# # # # # EigerStatus Tests # # # # #


@pytest.fixture
def eiger_settings() -> EigerSettings:
    return EigerSettings()


def test_eiger_settings_constructor():
    EigerSettings()


def test_eiger_settings_getitem(eiger_settings):
    value = eiger_settings["count_time"]["value"]

    assert 0.1 == value


def test_eiger_settings_get_element(eiger_settings):
    assert "Co" == eiger_settings.element


def test_eiger_settings_set_element(eiger_settings):
    eiger_settings["element"] = "Li"

    assert "Li" == eiger_settings.element
    assert KA_Energy["Li"].value == eiger_settings.photon_energy



from dataclasses import dataclass
from typing import Callable, Optional, Sequence

import pytest
from mock import AsyncMock, MagicMock, patch

from tickit.adapters.interpreters.command import CommandInterpreter
from tickit.adapters.interpreters.command.command_interpreter import Command
from tickit.core.adapter import Adapter

_GET_TYPE_HINTS = (
    "tickit.adapters.interpreters.command.command_interpreter.get_type_hints"
)


@pytest.fixture
def command_interpreter():
    return CommandInterpreter()


@pytest.fixture
def TestCommand():
    @dataclass(frozen=True)
    class TestCommand:
        command: bytes

        def __call__(self, func: Callable) -> Callable:
            setattr(func, "__command__", self)
            return func

        def parse(self, data: bytes) -> Optional[Sequence[str]]:
            if data == self.command:
                return ("arg1", "arg2")
            else:
                return None

    return TestCommand


@patch(
    _GET_TYPE_HINTS,
    lambda _: {"arg1": str, "arg2": str},
)
@pytest.mark.asyncio
async def test_command_interpreter_handle_calls_func_with_args(
    command_interpreter: CommandInterpreter,
):
    test_adapter = MagicMock(
        Adapter,
        test_method=AsyncMock(
            __command__=MagicMock(
                Command,
                interrupt=False,
                parse=MagicMock(return_value=("arg1", "arg2")),
            ),
        ),
    )
    await command_interpreter.handle(test_adapter, b"\x01")
    test_adapter.test_method.assert_awaited_once_with("arg1", "arg2")


@patch(
    _GET_TYPE_HINTS,
    lambda _: {"arg1": str, "arg2": str},
)
@pytest.mark.asyncio
async def test_command_interpreter_handle_returns_iterable_reply(
    command_interpreter: CommandInterpreter,
):
    reply = AsyncMock(__aiter__=MagicMock(), __anext__=AsyncMock())
    test_adapter = MagicMock(
        Adapter,
        test_method=AsyncMock(
            __command__=MagicMock(
                Command,
                interrupt=False,
                parse=MagicMock(return_value=("arg1", "arg2")),
            ),
            return_value=reply,
        ),
    )
    assert reply == (await command_interpreter.handle(test_adapter, b"\x01"))[0]


@patch(
    _GET_TYPE_HINTS,
    lambda _: {"arg1": str, "arg2": str},
)
@pytest.mark.asyncio
async def test_command_interpreter_handle_wraps_non_iterable_reply(
    command_interpreter: CommandInterpreter,
):
    reply = MagicMock("TestReply")
    test_adapter = MagicMock(
        Adapter,
        test_method=AsyncMock(
            __command__=MagicMock(
                Command,
                interrupt=False,
                parse=MagicMock(return_value=("arg1", "arg2")),
            ),
            return_value=reply,
        ),
    )
    assert (
        reply
        == await (await command_interpreter.handle(test_adapter, b"\x01"))[0]
        .__aiter__()
        .__anext__()
    )


@patch(
    _GET_TYPE_HINTS,
    lambda _: {"arg1": str, "arg2": str},
)
@pytest.mark.asyncio
@pytest.mark.parametrize("interrupt", [True, False])
async def test_command_interpreter_handle_returns_interupt(
    command_interpreter: CommandInterpreter, interrupt: bool
):
    test_adapter = MagicMock(
        Adapter,
        test_method=AsyncMock(
            __command__=MagicMock(
                Command,
                interrupt=interrupt,
                parse=MagicMock(return_value=("arg1", "arg2")),
            ),
            return_value="DummyReply",
        ),
    )
    assert interrupt == (await command_interpreter.handle(test_adapter, b"\x01"))[1]


@pytest.mark.asyncio
@pytest.mark.parametrize("interrupt", [True, False])
async def test_command_interpreter_handle_skips_unparsed(
    command_interpreter: CommandInterpreter, interrupt: bool
):
    test_adapter = MagicMock(
        Adapter,
        test_method=AsyncMock(
            __command__=MagicMock(
                Command, interrupt=interrupt, parse=MagicMock(return_value=None)
            ),
            return_value="DummyReply",
        ),
    )
    assert (
        b"Request does not match any known command"
        == await (
            await command_interpreter.handle(
                test_adapter, "TestCommand".encode("utf-8")
            )
        )[0]
        .__aiter__()
        .__anext__()
    )


@pytest.mark.asyncio
async def test_command_interpreter_handle_returns_message_for_no_commands(
    command_interpreter: CommandInterpreter,
):
    test_adapter = MagicMock(Adapter)
    assert (
        b"Request does not match any known command"
        == await (
            await command_interpreter.handle(
                test_adapter, "TestCommand".encode("utf-8")
            )
        )[0]
        .__aiter__()
        .__anext__()
    )



import subprocess
from typing import Iterable

import pytest
from click.testing import CliRunner, Result
from mock import Mock, patch
from mock.mock import create_autospec

from tickit import __version__
from tickit.cli import main
from tickit.core.components.component import ComponentConfig
from tickit.core.management.schedulers.master import MasterScheduler
from tickit.core.typedefs import ComponentID, ComponentPort, PortID


def test_cli_version():
    cmd = ["tickit", "--version"]
    assert (
        subprocess.check_output(cmd).decode().strip()
        == f"tickit, version {__version__}"
    )


@pytest.fixture
def patch_logging() -> Iterable[Mock]:
    with patch("tickit.cli.logging", autospec=True) as mock:
        yield mock


@pytest.fixture
def patch_run_all_forever() -> Iterable[Mock]:
    with patch("tickit.cli.run_all_forever", autospec=True) as mock:
        yield mock


@pytest.fixture
def patch_asyncio() -> Iterable[Mock]:
    with patch("tickit.cli.asyncio", autospec=True) as mock:
        yield mock


@pytest.fixture
def patch_read_configs() -> Iterable[Mock]:
    with patch("tickit.cli.read_configs", autospec=True) as mock:
        mock_config = create_autospec(ComponentConfig, instance=True)
        mock_config.name = "fake_device"
        mock_config.inputs = {
            PortID("42"),
            ComponentPort(ComponentID("foo"), PortID("24")),
        }
        mock.return_value = [mock_config]
        yield mock


def test_cli_set_loggging_level(
    patch_logging: Mock,
):
    runner: CliRunner = CliRunner()
    result: Result = runner.invoke(main, args=["--log-level", "INFO"])
    assert result.exit_code == 0
    patch_logging.basicConfig.assert_called_with(level="INFO")


def test_component_command(
    patch_run_all_forever: Mock,
    patch_read_configs: Mock,
):
    runner: CliRunner = CliRunner()

    result: Result = runner.invoke(
        main, args=["component", "fake_device", "path/to/fake_device.yaml"]
    )

    assert result.exit_code == 0
    patch_run_all_forever.assert_called_once()


@pytest.fixture
def patch_master_scheduler_run_forever_method() -> Iterable[Mock]:
    with patch.object(MasterScheduler, "run_forever", autospec=True) as mock:
        yield mock


def test_scheduler(
    patch_read_configs: Mock,
    patch_master_scheduler_run_forever_method: Mock,
):
    runner: CliRunner = CliRunner()

    result: Result = runner.invoke(main, args=["scheduler", "path/to/fake_device.yaml"])

    assert result.exit_code == 0
    patch_master_scheduler_run_forever_method.assert_awaited_once()


def test_all(
    patch_read_configs: Mock,
    patch_master_scheduler_run_forever_method: Mock,
):
    runner: CliRunner = CliRunner()

    result: Result = runner.invoke(main, args=["all", "path/to/fake_device.yaml"])

    assert result.exit_code == 0
    patch_master_scheduler_run_forever_method.assert_awaited_once()



import logging
from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict, Optional, Type, Union

from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer
from tickit.core.typedefs import (
    Changes,
    ComponentID,
    ComponentPort,
    Input,
    Interrupt,
    Output,
    PortID,
    SimTime,
)
from tickit.utils.configuration.configurable import as_tagged_union
from tickit.utils.topic_naming import input_topic, output_topic

LOGGER = logging.getLogger(__name__)


@dataclass  # type: ignore
@as_tagged_union
class Component:
    """An interface for types which implement stand-alone simulation components.

    An interface for types which implement stand-alone simulation components.
    Components define the top level building blocks of a tickit simulation (examples
    include the DeviceSimulation which host a device and corresponding adapter or a
    SystemSimulation which hosts a SlaveScheduler and internal Components).
    """

    name: ComponentID

    @abstractmethod
    async def run_forever(
        self, state_consumer: Type[StateConsumer], state_producer: Type[StateProducer]
    ) -> None:
        """An asynchronous method allowing indefinite running of core logic."""

    @abstractmethod
    async def on_tick(self, time: SimTime, changes: Changes):
        """An asynchronous method called whenever the component is to be updated.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            changes (Changes): A mapping of changed component inputs and their new
                values.
        """


@dataclass  # type: ignore
@as_tagged_union
class ComponentConfig:
    """A data container for component configuration.

    A data container for component configuration which acts as a named union of
    subclasses to facilitate automatic deserialization.
    """

    name: ComponentID
    inputs: Dict[PortID, ComponentPort]

    @abstractmethod
    def __call__(self) -> Component:
        """Create the component from the given config."""
        raise NotImplementedError(self)


class BaseComponent(Component):
    """A base class for compnents, implementing state interface related methods."""

    state_consumer: StateConsumer[Input]
    state_producer: StateProducer[Union[Interrupt, Output]]

    async def handle_input(self, input: Input):
        """Calls on_tick when an input is recieved.

        Args:
            input (Input): An immutable data container for Component inputs.
        """
        LOGGER.debug("{} got {}".format(self.name, input))
        await self.on_tick(input.time, input.changes)

    async def output(
        self,
        time: SimTime,
        changes: Changes,
        call_at: Optional[SimTime],
    ) -> None:
        """Constructs and sends an Output message to the component output topic.

        An asynchronous method which constructs an Output message tagged with the
        component name and sends it to the output topic of this component.

        Args:
            time (SimTime): The current time (in nanoseconds).
            changes (Changes): A mapping of the difference between the last observable
                state and the current observable state.
            call_at (Optional[SimTime]): The simulation time at which the component
                requests to be awoken.
        """
        await self.state_producer.produce(
            output_topic(self.name), Output(self.name, time, changes, call_at)
        )

    async def raise_interrupt(self) -> None:
        """Sends an Interrupt message to the component output topic.

        An asynchronous method whicb constructs an Interrupt message tagged with the
        component name and sends it to the output topic of this component.
        """
        await self.state_producer.produce(output_topic(self.name), Interrupt(self.name))

    async def run_forever(
        self, state_consumer: Type[StateConsumer], state_producer: Type[StateProducer]
    ) -> None:
        """Creates and configures a state consumer and state producer.

        An asynchronous method which creates a state consumer which is subscribed to
        the input topic of the component and calls back to handle_input, and a state
        producer to produce Interrupt or Output messages.
        """
        self.state_consumer = state_consumer(self.handle_input)
        await self.state_consumer.subscribe([input_topic(self.name)])
        self.state_producer = state_producer()

    @abstractmethod
    async def on_tick(self, time: SimTime, changes: Changes):
        """An abstract asynchronous method which implements the core logic of the
        component.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            changes (Changes): A mapping of changed component inputs and their new
                values.
        """
        raise NotImplementedError



from typing import AnyStr, AsyncIterable, Tuple

from tickit.adapters.interpreters.utils import wrap_as_async_iterable
from tickit.core.adapter import Adapter, Interpreter


class JoiningInterpreter(Interpreter[AnyStr]):
    """A wrapper for an interpreter that combines responses.

    An interpreter wrapper class that takes the wrapped interpreter's response(s) to a
    message and combines them into a single response.
    """

    def __init__(
        self,
        interpreter: Interpreter[AnyStr],
        response_delimiter: AnyStr,
    ) -> None:
        """A decorator for an interpreter that combines multiple responses into one.

        Args:
            interpreter (Interpreter): The interpreter responding to a message.
            response_delimiter (AnyStr): The delimiter separating the responses to the
                individual responses when they are combined into a single response
                message.

        """
        super().__init__()
        self.interpreter: Interpreter[AnyStr] = interpreter
        self.response_delimiter: AnyStr = response_delimiter

    async def _combine_responses(
        self, responses: AsyncIterable[AnyStr]
    ) -> AsyncIterable[AnyStr]:
        """Combines results from handling multiple messages.

        Takes the responses from when the wrapped interpreter handles multiple messages
        and returns an appropriate composite repsonse and interrrupt. The composite
        response is the concatentation of each of the individual responses, the
        composite interrupt is a logical inclusive 'or' of all of the individual
        responses.

        Args:
            responses (AsyncIterable[AnyStr]): an async iterable of reply messages from
                the wrapped class' handle() method.

        Returns:
            AsyncIterable[AnyStr]:
                An asynchronous iterable containing a single reply message.
        """
        response_list = [response async for response in responses]
        response = self.response_delimiter.join(response_list)
        return wrap_as_async_iterable(response)

    async def handle(
        self, adapter: Adapter, message: AnyStr
    ) -> Tuple[AsyncIterable[AnyStr], bool]:
        """Merges the responses from an interpreter into a single message.

        Individual responses to the message are combined into a single response and
            returned.

        Args:
            adapter (Adapter): The adapter in which the function should be executed.
            message: (AnyStr): The message to be handled.

        Returns:
            Tuple[AsyncIterable[Union[str, bytes]], bool]:
                A tuple of the asynchronous iterable of a single reply message and a
                flag indicating whether an interrupt should be raised by the adapter.
        """
        responses, interrupt = await self.interpreter.handle(adapter, message)
        resp = await self._combine_responses(responses)
        return resp, interrupt



from typing import AnyStr, AsyncIterable, List

import pytest
from mock import AsyncMock

from tickit.adapters.interpreters.wrappers import JoiningInterpreter


@pytest.mark.asyncio
async def test_joining_interpreter_concatenates_as_expected():
    async def multi_resp(msgs: List[AnyStr]) -> AsyncIterable[AnyStr]:
        for msg in msgs:
            yield msg

    mock_interpreter = AsyncMock()
    mock_interpreter.handle.return_value = multi_resp(["one", "two", "three"]), True
    joining_interpreter = JoiningInterpreter(mock_interpreter, response_delimiter="-")
    result, _ = await joining_interpreter.handle(AsyncMock(), "test")
    assert [msg async for msg in result] == ["one-two-three"]



from dataclasses import dataclass
from typing import AnyStr, Awaitable, Callable, Generic

from aiohttp import web
from aiohttp.web_response import StreamResponse
from aiohttp.web_routedef import RouteDef


@dataclass(frozen=True)
class HTTPEndpoint(Generic[AnyStr]):
    """A decorator to register a device adapter method as a HTTP Endpoint.

    Args:
        url (str): The URL that will point to a specific endpoint.
        method (str): The method to use when using this endpoint.
        name (str): The name of the route.
        include_json (bool): A flag to indicate whether the route should include json.
        interrupt (bool): A flag indicating whether calling of the method should
            raise an adapter interrupt. Defaults to False.

    Returns:
        Callable:
            A decorator which registers the adapter method as an endpoint.
    """

    url: str
    method: str
    include_json: bool = False
    interrupt: bool = False

    # Type signature can become more specific if support is dropped for
    # Python 3.7, see https://github.com/python/mypy/issues/708
    def __call__(self, func: Callable) -> Callable:
        """A decorator which registers the adapter method as an endpoint.

        Args:
            func (Callable): The adapter method to be registered as an endpoint.

        Returns:
            Callable: The registered adapter endpoint.
        """
        setattr(func, "__endpoint__", self)
        return func

    def define(
        self, func: Callable[[web.Request], Awaitable[StreamResponse]]
    ) -> RouteDef:
        """Performs the construction of the endpoint RouteDef for the HTTP Server.

        A method which performs the construction of the route definition of the method,
         URL and handler function for the endpoint, to then return to the HTTP Server.

        Args:
            func (Callable): The handler funcion to be attached to the route
            definition for the endpoint.

        Returns:
            RouteDef: The route definition for the endpoint.
        """
        return RouteDef(self.method, self.url, func, {})

    @classmethod
    def get(
        cls, url: str, include_json: bool = False, interrupt: bool = False
    ) -> "HTTPEndpoint":
        """Method for the HTTPEndpoint that sets the request method to "GET".

        Returns:
            cls: The class of HTTPEndpoint with the "GET" request method.
        """
        return cls(url, "GET", include_json, interrupt)

    @classmethod
    def put(
        cls, url: str, include_json: bool = False, interrupt: bool = False
    ) -> "HTTPEndpoint":
        """Method for the HTTPEndpoint that sets the request method to "PUT".

        Returns:
            cls: The class of HTTPEndpoint with the "PUT" request method.
        """
        return cls(url, "PUT", include_json, interrupt)



import asyncio
import logging
import struct
from typing import Optional

import numpy as np
import pytest

from tickit.core.device import DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.devices.cryostream.cryostream import CryostreamDevice
from tickit.devices.cryostream.states import PhaseIds
from tickit.devices.cryostream.status import ExtendedStatus

# # # # # Cryostream Tests # # # # #


@pytest.fixture
def cryostream() -> CryostreamDevice:
    return CryostreamDevice()


def test_cryostream_constructor():
    CryostreamDevice()


def test_cryostream_update_hold(cryostream):
    starting_temperature = cryostream.gas_temp
    device_update: DeviceUpdate = cryostream.update(time=SimTime(1e9), inputs={})
    device_update.outputs["temperature"] == starting_temperature


@pytest.mark.asyncio
async def test_cryostream_update_cool(cryostream: CryostreamDevice):
    starting_temperature = cryostream.gas_temp
    target_temperature = starting_temperature - 50
    await cryostream.cool(target_temperature)

    time = SimTime(0)
    device_update: DeviceUpdate
    while cryostream.phase_id != PhaseIds.HOLD.value:
        logging.info(f"Running time step: {time}")
        device_update = cryostream.update(time, inputs={})
        time_update: Optional[SimTime] = device_update.call_at

        if time_update is None:
            time = SimTime(int(time) + int(1e9))
            continue
        else:
            time = time_update

    max_diff = 10
    margin_of_error = np.array([+max_diff, -max_diff])
    assert any(
        (device_update.outputs["temperature"] + margin_of_error) == target_temperature
    )


@pytest.mark.asyncio
async def test_cryostream_update_end(cryostream: CryostreamDevice):
    starting_temperature = cryostream.default_temp_shutdown - 100
    cryostream.gas_temp = starting_temperature
    await cryostream.end(cryostream.default_ramp_rate)
    assert cryostream.phase_id == PhaseIds.RAMP.value
    assert cryostream.gas_flow == 5

    time = SimTime(0)
    device_update: DeviceUpdate
    while cryostream.phase_id != PhaseIds.HOLD.value:
        logging.info(f"Running time step: {time}")
        device_update = cryostream.update(time, inputs={})
        time_update: Optional[SimTime] = device_update.call_at

        if time_update is None:
            continue
        else:
            time = time_update

    max_diff = 10
    target = cryostream.default_temp_shutdown
    assert (
        target - max_diff <= device_update.outputs["temperature"] <= target + max_diff
    )


@pytest.mark.asyncio
async def test_cryostream_update_plat(cryostream: CryostreamDevice):
    starting_temperature = cryostream.gas_temp
    await cryostream.plat(5)
    assert cryostream.phase_id == PhaseIds.PLAT.value

    time = SimTime(0)
    time_update: Optional[SimTime] = time
    device_update: DeviceUpdate
    while time_update is not None:
        logging.info(f"Running time step: {time}")
        device_update = cryostream.update(time, inputs={})
        time_update = device_update.call_at

        if time_update is None:
            continue
        else:
            time = time_update

    assert device_update.outputs["temperature"] == starting_temperature


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "tickit_task", ["examples/configs/cryo-tcp.yaml"], indirect=True
)
async def test_cryostream_system(tickit_task):
    reader, writer = await asyncio.open_connection("localhost", 25565)

    async def write(data: bytes):
        writer.write(data)
        await writer.drain()

    async def get_status() -> ExtendedStatus:
        return ExtendedStatus.from_packed(await reader.read(42))

    # Check we start at 300K
    status = await get_status()
    assert status.gas_temp == 30000
    assert status.turbo_mode == 0
    # Start a ramp
    await write(b"\x06\x0b" + struct.pack(">HH", 360, 30030))
    # Check we go to to 300.3K, eventually
    status = await get_status()
    assert status.gas_temp == pytest.approx(30015, rel=5)
    status = await get_status()
    assert status.gas_temp == pytest.approx(30025, rel=5)
    status = await get_status()
    assert status.gas_temp == pytest.approx(30030, rel=5)
    # Turbo on
    await write(b"\x03\x14\x01")
    status = await get_status()
    assert status.gas_temp == pytest.approx(30030, rel=5)
    assert status.turbo_mode == 1
    # Cool to 300K
    await write(b"\x04\x0e" + struct.pack(">H", 30000))
    status = await get_status()
    assert status.gas_temp == pytest.approx(30000, rel=5)



from typing import AnyStr, AsyncIterable, Tuple

from tickit.core.adapter import Adapter, Interpreter


class BeheadingInterpreter(Interpreter[AnyStr]):
    """A wrapper for an interpreter which strips a header from a message.

    An interpreter wrapper that takes a message, strips off a header of a fixed size
    and passes the stripped message on to the wrapped interpreter
    """

    def __init__(self, interpreter: Interpreter[AnyStr], header_size: int) -> None:
        """A wrapper for an interpreter which strips a header from a message.

        Args:
            interpreter (Interpreter): The interpreter the message is passed on to
                after the header is stripped.
            header_size (int): The number of characters in the header.
        """
        super().__init__()
        self.interpreter: Interpreter[AnyStr] = interpreter
        self.header_size: int = header_size

    async def handle(
        self, adapter: Adapter, message: AnyStr
    ) -> Tuple[AsyncIterable[AnyStr], bool]:
        """Removes a header from the start of a message, and passes it on to an
        interpreter.

        Args:
            adapter (Adapter): The adapter in which the function should be executed
            message: (AnyStr): The handled message, of which the header is removed.

        Returns:
            Tuple[AsyncIterable[Union[str, bytes]], bool]:
                A tuple of the asynchronous iterable of reply messages and a flag
                indicating whether an interrupt should be raised by the adapter.
        """
        message = message[self.header_size :]
        return await self.interpreter.handle(adapter, message)



from typing import Dict, Hashable, Mapping, Set

import pytest
from immutables import Map

from tickit.core.components.component import ComponentConfig
from tickit.core.management.event_router import EventRouter, InverseWiring, Wiring
from tickit.core.typedefs import Changes, ComponentID, ComponentPort, PortID


@pytest.fixture
def wiring_struct():
    return {
        ComponentID("Out1"): {
            PortID("Out1>1"): {ComponentPort(ComponentID("Mid1"), PortID("Mid1<1"))}
        },
        ComponentID("Out2"): {
            PortID("Out2>1"): {
                ComponentPort(ComponentID("In1"), PortID("In1<2")),
                ComponentPort(ComponentID("Mid1"), PortID("Mid1<2")),
            }
        },
        ComponentID("Mid1"): {
            PortID("Mid1>1"): {ComponentPort(ComponentID("In1"), PortID("In1<1"))}
        },
        ComponentID("In1"): {},
    }


@pytest.fixture
def inverse_wiring_struct():
    return {
        ComponentID("Out1"): {},
        ComponentID("Out2"): {},
        ComponentID("Mid1"): {
            PortID("Mid1<1"): ComponentPort(ComponentID("Out1"), PortID("Out1>1")),
            PortID("Mid1<2"): ComponentPort(ComponentID("Out2"), PortID("Out2>1")),
        },
        ComponentID("In1"): {
            PortID("In1<1"): ComponentPort(ComponentID("Mid1"), PortID("Mid1>1")),
            PortID("In1<2"): ComponentPort(ComponentID("Out2"), PortID("Out2>1")),
        },
    }


@pytest.fixture
def component_configs_list():
    return [
        ComponentConfig(ComponentID("Out1"), dict()),
        ComponentConfig(ComponentID("Out2"), dict()),
        ComponentConfig(
            ComponentID("Mid1"),
            {
                PortID("Mid1<1"): ComponentPort(ComponentID("Out1"), PortID("Out1>1")),
                PortID("Mid1<2"): ComponentPort(ComponentID("Out2"), PortID("Out2>1")),
            },
        ),
        ComponentConfig(
            ComponentID("In1"),
            {
                PortID("In1<1"): ComponentPort(ComponentID("Mid1"), PortID("Mid1>1")),
                PortID("In1<2"): ComponentPort(ComponentID("Out2"), PortID("Out2>1")),
            },
        ),
    ]


@pytest.fixture
def wiring(wiring_struct):
    return Wiring(wiring_struct)


@pytest.fixture
def event_router(wiring):
    return EventRouter(wiring)


def test_wiring_unknown_out_dev(wiring_struct):
    wiring = Wiring(wiring_struct)
    assert dict() == wiring["Out3"]


def test_wiring_unknown_out_io(wiring_struct):
    wiring = Wiring(wiring_struct)
    assert set() == wiring["Out1"]["Out1>2"]


def test_wiring_from_inverse_equal(wiring_struct, inverse_wiring_struct):
    assert Wiring(wiring_struct) == Wiring.from_inverse_wiring(inverse_wiring_struct)


def test_inverse_wiring_unknown_in_dev(inverse_wiring_struct):
    inverse_wiring = InverseWiring(inverse_wiring_struct)
    assert dict() == inverse_wiring["In2"]


def test_inverse_wiring_unknown_in_io(inverse_wiring_struct):
    inverse_wiring = InverseWiring(inverse_wiring_struct)
    with pytest.raises(KeyError):
        inverse_wiring["In1"]["In1<3"]


def test_inverse_wiring_from_wiring_equal(inverse_wiring_struct, wiring_struct):
    assert InverseWiring(inverse_wiring_struct) == InverseWiring.from_wiring(
        wiring_struct
    )


def test_inverse_wiring_from_component_configs_equal(
    inverse_wiring_struct, component_configs_list
):
    assert InverseWiring(inverse_wiring_struct) == InverseWiring.from_component_configs(
        component_configs_list
    )


def test_event_router_wiring_from_wiring(wiring: Wiring):
    event_router = EventRouter(wiring)
    assert wiring == event_router.wiring


def test_event_router_wiring_from_inverse(wiring: Wiring):
    inverse_wiring = InverseWiring.from_wiring(wiring)
    event_router = EventRouter(inverse_wiring)
    assert wiring == event_router.wiring


def test_event_router_components(event_router: EventRouter):
    assert {"Out1", "Out2", "Mid1", "In1"} == event_router.components


def test_event_router_input_components(event_router: EventRouter):
    assert {"Mid1", "In1"} == event_router.input_components


def test_event_router_output_components(event_router: EventRouter):
    assert {"Out1", "Out2", "Mid1"} == event_router.output_components


def test_event_router_component_tree(event_router: EventRouter):
    assert {
        "In1": set(),
        "Out1": {"Mid1"},
        "Out2": {"Mid1", "In1"},
        "Mid1": {"In1"},
    } == event_router.component_tree


def test_event_router_inverse_component_tree(event_router: EventRouter):
    assert {
        "Out1": set(),
        "Out2": set(),
        "Mid1": {"Out1", "Out2"},
        "In1": {"Mid1", "Out2"},
    } == event_router.inverse_component_tree


@pytest.mark.parametrize(
    ["root", "expected"],
    [
        ("Out1", {"Out1", "Mid1", "In1"}),
        ("Out2", {"Out2", "Mid1", "In1"}),
        ("Mid1", {"Mid1", "In1"}),
        ("In1", {"In1"}),
    ],
)
def test_event_router_dependants(
    event_router: EventRouter, root: ComponentID, expected: Set[ComponentID]
):
    assert expected == event_router.dependants(root)


@pytest.mark.parametrize(
    ["source", "changes", "expected"],
    [
        (
            ComponentID("Out1"),
            Changes(Map({PortID("Out1>1"): 42})),
            {ComponentID("Mid1"): {PortID("Mid1<1"): 42}},
        )
    ],
)
def test_event_router_route(
    event_router: EventRouter,
    source: ComponentID,
    changes: Mapping[PortID, Hashable],
    expected: Dict[ComponentID, Dict[PortID, object]],
):
    assert expected == event_router.route(source, changes)



import logging
from dataclasses import dataclass
from typing import Any

from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.compat.typing_compat import TypedDict

LOGGER = logging.getLogger(__name__)


class SourceDevice(Device):
    """A simple device which produces a pre-configured value."""

    #: An empty typed mapping of device inputs
    Inputs: TypedDict = TypedDict("Inputs", {})
    #: A typed mapping containing the 'value' output value
    Outputs: TypedDict = TypedDict("Outputs", {"value": Any})

    def __init__(self, value: Any) -> None:
        """A constructor of the source, which takes the pre-configured output value.

        Args:
            value (Any): A pre-configured output value.
        """
        self.value = value

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """The update method which produces the pre-configured output value.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the pre-configured value, and
                never requests a callback.
        """
        LOGGER.debug("Sourced {}".format(self.value))
        return DeviceUpdate(SourceDevice.Outputs(value=self.value), None)


@dataclass
class Source(ComponentConfig):
    """Source of a fixed value."""

    value: Any

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=SourceDevice(self.value),
        )



import sys

if sys.version_info >= (3, 8):
    from typing import Protocol, TypedDict, runtime_checkable
elif sys.version_info >= (3, 5):
    from typing_extensions import Protocol, TypedDict, runtime_checkable

__all__ = ["Protocol", "TypedDict", "runtime_checkable"]



import asyncio
from time import time_ns
from typing import Type, Union

from tickit.core.management.event_router import InverseWiring, Wiring
from tickit.core.management.schedulers.base import BaseScheduler
from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer
from tickit.core.typedefs import ComponentID, SimTime


class MasterScheduler(BaseScheduler):
    """A master scheduler which orchestrates the running of a tickit simulation."""

    def __init__(
        self,
        wiring: Union[Wiring, InverseWiring],
        state_consumer: Type[StateConsumer],
        state_producer: Type[StateProducer],
        initial_time: int = 0,
        simulation_speed: float = 1.0,
    ) -> None:
        """A master scheduler constructor  which stores values for reference.

        Args:
            wiring (Union[Wiring, InverseWiring]): A wiring or inverse wiring object
                representing the connections between components in the system.
            state_consumer (Type[StateConsumer]): The state consumer class to be used
                by the component.
            state_producer (Type[StateProducer]): The state producer class to be used
                by the component.
            initial_time (int): The initial time of the simulation (in nanoseconds).
                Defaults to 0.
            simulation_speed (float): The rate at which simulation time progresses with
                respect to real world time. Defaults to 1.0.
        """
        super().__init__(wiring, state_consumer, state_producer)

        self._initial_time = SimTime(initial_time)
        self.simulation_speed = simulation_speed

    async def setup(self) -> None:
        """Performs base setup and creates an awaitable flag to indicate new wakeups."""
        await super().setup()
        self.new_wakeup: asyncio.Event = asyncio.Event()

    def add_wakeup(self, component: ComponentID, when: SimTime) -> None:
        """Adds a wakeup to the priority queue and sets an awaitable flag.

        A method which adds a wakeup to the priority queue and sets a awaitable flag
        indicating that a new wakeup has been added.

        Args:
            component (ComponentID): The component which should be updated.
            when (SimTime): The simulation time at which the update should occur.
        """
        super().add_wakeup(component, when)
        self.new_wakeup.set()

    async def run_forever(self) -> None:
        """Performs an intial tick then continiously schedules ticks according to
        wakeups.

        An asynchronous method which initially performs setup and an initial tick in
        which all components are updated, subsequently ticks are performed as requested
        by components of the simulation according to the simulation speed.
        """
        await self.setup()
        await self._do_initial_tick()
        while True:
            await self._do_tick()

    async def _do_initial_tick(self):
        """Performs the initial tick of the system."""
        await self.ticker(
            self._initial_time,
            self.ticker.components,
        )
        self.last_time = time_ns()

    async def _do_tick(self):
        """Continuously schedules ticks according to wakeups."""
        if not self.wakeups:
            await self.new_wakeup.wait()
        components, when = self.get_first_wakeups()
        assert when is not None
        self.new_wakeup.clear()

        current: asyncio.Future = asyncio.sleep(self.sleep_time(when))
        new = asyncio.create_task(self.new_wakeup.wait())
        which, _ = await asyncio.wait(
            [current, new], return_when=asyncio.tasks.FIRST_COMPLETED
        )

        if new in which:
            return

        for component in components:
            del self.wakeups[component]
        await self.ticker(when, {component for component in components})
        self.last_time = time_ns()

    async def schedule_interrupt(self, source: ComponentID) -> None:
        """Schedules the interrupt of a component immediately.

        An asynchronous method which implements the superclass abstract, an interrupt
        is sheduled as an immediate wakeup. This is achieved by giving the wakeup a
        simulation time equal to the simulation time of the last tick plus the real
        world time which has passed, adjusted by the simulation speed.

        Args:
            source (ComponentID): The source component which should be updated.
        """
        self.add_wakeup(
            source,
            SimTime(
                self.ticker.time
                + int((time_ns() - self.last_time) * self.simulation_speed)
            ),
        )

    def sleep_time(self, when: SimTime) -> float:
        """Computes the real world time until a specified simulation time.

        Args:
            when (SimTime): The simulation time to be reached.

        Returns:
            float: The real world duration before the simulation time is reached.
        """
        return (
            (when - self.ticker.time) / self.simulation_speed
            - (time_ns() - self.last_time)
        ) / 1e9



import asyncio
from typing import Any, Awaitable, Callable, Iterable

import pytest
from immutables import Map
from mock import AsyncMock, Mock, patch
from mock.mock import create_autospec

from tickit.core.components.component import Component
from tickit.core.components.system_simulation import (
    SystemSimulation,
    SystemSimulationComponent,
)
from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer
from tickit.core.typedefs import (
    Changes,
    ComponentID,
    ComponentPort,
    Output,
    PortID,
    SimTime,
)
from tickit.utils.topic_naming import output_topic


class MockStateConsumer:
    @staticmethod
    def __call__(callback: Callable[[Any], Awaitable[None]]) -> None:
        return create_autospec(StateConsumer)


class MockStateProducer:
    @staticmethod
    def __call__():
        return create_autospec(StateProducer)


@pytest.fixture
def patch_scheduler() -> Iterable[Mock]:
    spec = "tickit.core.components.system_simulation.SlaveScheduler"
    with patch(spec, autospec=True) as mock:

        def on_tick(time, changes):
            pass

        mock.return_value = AsyncMock()
        dummy_output = (Changes(Map({PortID("84"): 84})), 3)
        mock.return_value.on_tick = AsyncMock(spec=on_tick, return_value=dummy_output)
        mock.return_value.setup
        yield mock


@pytest.fixture
def system_simulation(patch_scheduler) -> Component:
    system_simulation_config = SystemSimulation(
        name=ComponentID("test_system_simulation"),
        inputs={PortID("24"): ComponentPort(ComponentID("25"), PortID("23"))},
        components=[],
        expose={PortID("42"): ComponentPort(ComponentID("43"), PortID("44"))},
    )

    return system_simulation_config()


def test_system_simulation_constructor(system_simulation: SystemSimulation):
    pass


@pytest.fixture
def patch_asyncio() -> Iterable[Mock]:
    with patch(
        "tickit.core.components.system_simulation.asyncio", autospec=True
    ) as mock:
        yield mock


@pytest.fixture
def patch_run_all() -> Iterable[Mock]:
    with patch(
        "tickit.core.components.system_simulation.run_all", autospec=True
    ) as mock:
        mock.return_value = [asyncio.sleep(0)]
        yield mock


@pytest.mark.asyncio
async def test_system_simulation_methods(
    system_simulation: SystemSimulationComponent,
    patch_run_all: Mock,
):
    """Test the 'run_forever' and 'on_tick' methods.

    The 'on_tick' method depends on 'run_forever' being called first. Therefore tests
    for both methods have been bundled together. The test succeeds if the mock state
    interfaces of the system simulation are awaited once with the correct parameters.
    """
    await system_simulation.run_forever(
        MockStateConsumer(), MockStateProducer()  # type: ignore
    )

    system_simulation.state_consumer.subscribe.assert_awaited_once()  # type: ignore
    time = SimTime(0)
    (
        expected_changes,
        expected_call_back,
    ) = system_simulation.scheduler.on_tick.return_value  # type: ignore
    await system_simulation.on_tick(time, expected_changes)

    system_simulation.state_producer.produce.assert_awaited_once_with(  # type: ignore
        output_topic(system_simulation.name),
        Output(system_simulation.name, time, expected_changes, expected_call_back),
    )



from dataclasses import dataclass

from tickit.adapters.composed import ComposedAdapter
from tickit.adapters.interpreters.command import CommandInterpreter, RegexCommand
from tickit.adapters.servers.tcp import TcpServer
from tickit.core.adapter import Server
from tickit.core.components.component import Component, ComponentConfig
from tickit.core.components.device_simulation import DeviceSimulation
from tickit.core.device import Device, DeviceUpdate
from tickit.core.typedefs import SimTime
from tickit.utils.byte_format import ByteFormat
from tickit.utils.compat.typing_compat import TypedDict


class IoBoxDevice(Device):
    """An isolated toy device which stores a message.

    The device takes no inputs and produces no outputs. It interacts exclusively with
    an adapter which sets incoming messages to it and reads stored messages from it.
    """

    #: An empty typed mapping of device inputs
    Inputs: TypedDict = TypedDict("Inputs", {})
    #: A typed mapping containing the current output value
    Outputs: TypedDict = TypedDict("Outputs", {})

    def __init__(self, inital_value: str = "Hello") -> None:
        """The IoBox constructor, sets intial message value."""
        self.message = inital_value

    def update(self, time: SimTime, inputs: Inputs) -> DeviceUpdate[Outputs]:
        """The update method which produces an output mapping containing the observed
        value.

        For this device the DeviceUpdate produces no outputs.

        Args:
            time (SimTime): The current simulation time (in nanoseconds).
            inputs (State): A mapping of inputs to the device and their values.

        Returns:
            DeviceUpdate[Outputs]:
                The produced update event which contains the observed value, in this
                case nothing is provided. The device never requests a callback.
        """
        return DeviceUpdate(IoBoxDevice.Outputs(), None)


class IOBoxAdapater(ComposedAdapter):
    """A Composed adapter for setting and getting the device property."""

    device: IoBoxDevice

    def __init__(
        self,
        server: Server,
    ) -> None:
        """Constructor of the IoBoc Adapter, builds the configured server.

        Args:
            server (Server): The immutable data container used to configure a
                server.
        """
        super().__init__(
            server,
            CommandInterpreter(),
        )

    @RegexCommand(r"m=([a-zA-Z0-9_!.?-]+)", interrupt=True, format="utf-8")
    async def set_message(self, value: str) -> None:
        """Regex string command which sets the value of the message.

        Args:
            value (str): The new value of the message.
        """
        self.device.message = value

    @RegexCommand(r"m\?", format="utf-8")
    async def get_message(self) -> bytes:
        """Regex string command which returns the utf-8 encoded value of the message.

        Returns:
            bytes: The utf-8 encoded value of the message.
        """
        return str(self.device.message).encode("utf-8")


@dataclass
class IoBox(ComponentConfig):
    """IO Box you can talk to over TCP."""

    host: str = "localhost"
    port: int = 25565
    format: ByteFormat = ByteFormat(b"%b\r\n")

    def __call__(self) -> Component:  # noqa: D102
        return DeviceSimulation(
            name=self.name,
            device=IoBoxDevice(),
            adapters=[IOBoxAdapater(TcpServer(self.host, self.port, self.format))],
        )



from dataclasses import dataclass, field, fields
from enum import Enum
from typing import Any, List

from .eiger_schema import (
    AccessMode,
    field_config,
    ro_float,
    ro_str,
    rw_bool,
    rw_float,
    rw_int,
    rw_str,
)

FRAME_WIDTH: int = 4148
FRAME_HEIGHT: int = 4362


class KA_Energy(Enum):
    """Possible element K-alpha energies for samples."""

    Li: float = 54.3
    Be: float = 108.5
    B: float = 183.3
    C: float = 277.0
    N: float = 392.4
    O: float = 524.9
    F: float = 676.8
    Ne: float = 848.6
    Na: float = 1040.98
    Mg: float = 1253.6
    Al: float = 1486.7
    Si: float = 1739.98
    P: float = 2013.7
    S: float = 2307.84
    Cl: float = 2622.39
    Ar: float = 2957.7
    K: float = 3313.8
    Ca: float = 3691.68
    Sc: float = 4090.6
    Ti: float = 4510.84
    V: float = 4952.2
    Cr: float = 5414.72
    Mn: float = 5898.75
    Fe: float = 6403.84
    Co: float = 6930.32
    Ni: float = 7478.15
    Cu: float = 8047.78
    Zn: float = 8638.86


@dataclass
class EigerSettings:
    """A data container for Eiger device configuration."""

    auto_summation: bool = field(default=True, metadata=rw_bool())
    beam_center_x: float = field(default=0.0, metadata=rw_float())
    beam_center_y: float = field(default=0.0, metadata=rw_float())
    bit_depth_image: int = field(default=16, metadata=rw_int())
    bit_depth_readout: int = field(default=16, metadata=rw_int())
    chi_increment: float = field(default=0.0, metadata=rw_float())
    chi_start: float = field(default=0.0, metadata=rw_float())
    compression: str = field(
        default="bslz4", metadata=rw_str(allowed_values=["bslz4", "lz4"])
    )
    count_time: float = field(default=0.1, metadata=rw_float())
    countrate_correction_applied: bool = field(default=True, metadata=rw_bool())
    countrate_correction_count_cutoff: int = field(default=1000, metadata=rw_int())
    data_collection_date: str = field(
        default="2021-30-09T16:30:00.000-01:00", metadata=ro_str()
    )
    description: str = field(
        default="Simulated Eiger X 16M Detector", metadata=ro_str()
    )
    detector_distance: float = field(default=2.0, metadata=rw_float())
    detector_number: str = field(default="EIGERSIM001", metadata=ro_str())
    detector_readout_time: float = field(default=0.01, metadata=rw_float())
    element: str = field(
        default="Co", metadata=rw_str(allowed_values=[e.name for e in KA_Energy])
    )
    flatfield: List[List[float]] = field(
        default_factory=lambda: [[]],
        metadata=field_config(value_type=AccessMode.FLOAT_GRID),
    )
    flatfield_correction_applied: bool = field(default=True, metadata=rw_bool())
    frame_time: float = field(default=0.12, metadata=rw_float())
    kappa_increment: float = field(default=0.0, metadata=rw_float())
    kappa_start: float = field(default=0.0, metadata=rw_float())
    nimages: int = field(default=1, metadata=rw_int())
    ntrigger: int = field(default=1, metadata=rw_int())
    number_of_excuded_pixels: int = field(default=0, metadata=rw_int())
    omega_increment: float = field(default=0.0, metadata=rw_float())
    omega_start: float = field(default=0.0, metadata=rw_float())
    phi_increment: float = field(default=0.0, metadata=rw_float())
    phi_start: float = field(default=0.0, metadata=rw_float())
    photon_energy: float = field(default=6930.32, metadata=rw_float())
    pixel_mask: List[List[int]] = field(
        default_factory=lambda: [[]],
        metadata=field_config(value_type=AccessMode.UINT_GRID),
    )
    pixel_mask_applied: bool = field(default=False, metadata=rw_bool())
    roi_mode: str = field(
        default="disabled", metadata=rw_str(allowed_values=["disabled", "4M"])
    )
    sensor_material: str = field(default="Silicon", metadata=ro_str())
    sensor_thickness: float = field(default=0.01, metadata=ro_float())
    software_version: str = field(default="0.1.0", metadata=ro_str())
    threshold_energy: float = field(default=4020.5, metadata=rw_float())
    trigger_mode: str = field(
        default="exts", metadata=rw_str(allowed_values=["exts", "ints", "exte", "inte"])
    )
    two_theta_increment: float = field(default=0.0, metadata=rw_float())
    two_theta_start: float = field(default=0.0, metadata=rw_float())
    wavelength: float = field(default=1e-9, metadata=rw_float())
    x_pixel_size: float = field(default=0.01, metadata=ro_float())
    x_pixels_in_detector: int = field(default=FRAME_WIDTH, metadata=rw_int())
    y_pixel_size: float = field(default=0.01, metadata=ro_float())
    y_pixels_in_detector: int = field(default=FRAME_HEIGHT, metadata=rw_int())

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        f = {}
        for field_ in fields(self):
            f[field_.name] = {
                "value": vars(self)[field_.name],
                "metadata": field_.metadata,
            }
        return f[key]

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: D105
        self.__dict__[key] = value

        if key == "element":
            self.photon_energy = getattr(KA_Energy, value).value
            self.threshold_energy = 0.5 * self.photon_energy



import pytest

from tickit.adapters.interpreters.utils import (
    wrap_as_async_iterable,
    wrap_messages_as_async_iterable,
)


@pytest.mark.asyncio
async def test_wrap_list_correctly():
    messages = ["Hello", "World"]
    wrapped = wrap_messages_as_async_iterable(messages)
    assert "Hello" == await wrapped.__anext__()
    assert "World" == await wrapped.__anext__()
    with pytest.raises(StopAsyncIteration):
        await wrapped.__anext__()


@pytest.mark.asyncio
async def test_wrap_message_correctly():
    message = b"Hello World"
    wrapped = wrap_as_async_iterable(message)
    assert b"Hello World" == await wrapped.__anext__()
    with pytest.raises(StopAsyncIteration):
        await wrapped.__anext__()



from dataclasses import dataclass, field, fields
from typing import Any, List

from tickit.devices.eiger.eiger_schema import AccessMode


@dataclass
class MonitorStatus:
    """Eiger monitor status taken from the API spec."""

    error: List[str] = field(
        default_factory=lambda: [],
        metadata=dict(
            value=[], value_type=AccessMode.LIST_STR, access_mode=AccessMode.READ_ONLY
        ),
    )

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        f = {}
        for field_ in fields(self):
            f[field_.name] = {
                "value": vars(self)[field_.name],
                "metadata": field_.metadata,
            }
        return f[key]



import asyncio
import logging
from dataclasses import dataclass
from inspect import getmembers
from typing import Iterable

from aiohttp import web
from aiohttp.web_routedef import RouteDef

from tickit.adapters.interpreters.endpoints.http_endpoint import HTTPEndpoint
from tickit.core.adapter import Adapter, RaiseInterrupt
from tickit.core.device import Device

LOGGER = logging.getLogger(__name__)


@dataclass
class HTTPAdapter(Adapter):
    """An adapter implementation which delegates to a server and sets up endpoints.

    An adapter implementation which delegates the hosting of an http requests to a
    server and sets up the endpoints for said server.
    """

    host: str = "localhost"
    port: int = 8080

    async def run_forever(
        self, device: Device, raise_interrupt: RaiseInterrupt
    ) -> None:
        """Runs the server continously."""
        await super().run_forever(device, raise_interrupt)

        await self._start_server()

        try:
            await asyncio.Event().wait()
        finally:
            # TODO: This doesn't work yet due to asyncio's own exception handler
            await self.app.shutdown()
            await self.app.cleanup()

    async def _start_server(self):
        LOGGER.debug(f"Starting HTTP server... {self}")
        self.app = web.Application()
        self.app.add_routes(list(self.endpoints()))
        runner = web.AppRunner(self.app)
        await runner.setup()
        site = web.TCPSite(runner, host=self.host, port=self.port)
        await site.start()

    def endpoints(self) -> Iterable[RouteDef]:
        """Returns list of endpoints.

        Fetches the defined HTTP endpoints in the device adapter, parses them and
        then yields them.

        Returns:
            Iterable[HTTPEndpoint]: The list of defined endpoints

        Yields:
            Iterator[Iterable[HTTPEndpoint]]: The iterator of the defined endpoints
        """
        for _, func in getmembers(self):
            endpoint = getattr(func, "__endpoint__", None)  # type: ignore
            if endpoint is not None and isinstance(endpoint, HTTPEndpoint):
                yield endpoint.define(func)



from typing import Any, Dict


class Singleton(type):
    """A singlton metaclass, which allows only one instance of derived types."""

    _instances: Dict["Singleton", "Singleton"] = {}

    def __call__(self, *args: Any, **kwargs: Any) -> "Singleton":
        """Return the class instance, if uninstantiated create an instance."""
        if self not in self._instances:
            self._instances[self] = super(Singleton, self).__call__(*args, **kwargs)
        return self._instances[self]



from importlib import import_module
from typing import Any, Dict, List, Optional, Type

import yaml
from apischema import deserialize
from apischema.conversions import AnyConversion, Conversion
from apischema.conversions.conversions import Conversion
from apischema.conversions.converters import default_deserialization

from tickit.core.components.component import ComponentConfig
from tickit.utils.configuration.configurable import is_tagged_union


def importing_conversion(typ: Type) -> Optional[AnyConversion]:
    """Create a conversion that imports the module of a ComponentConfig.

    When a ComponentConfig is requested from a dict, take its fully qualified
    name from the tagged union dict and import it before deserializing it
    """
    if is_tagged_union[typ]:

        def conversion(d: Dict[str, Any]):
            # We can't use the deserialization union above as the classes
            # haven't been imported so won't appear in __subclasses__, so use a
            # single element dict instead
            assert len(d) == 1, d
            fullname, args = list(d.items())[0]
            pkg, clsname = fullname.rsplit(".", maxsplit=1)
            cls = getattr(import_module(pkg), clsname)
            return deserialize(cls, args, default_conversion=importing_conversion)

        return Conversion(conversion, source=dict, target=typ)

    return default_deserialization(typ)


def read_configs(config_path) -> List[ComponentConfig]:
    """A utility function which reads and deserializes configs.

    A utility function which reads config files, performs yaml deserialization,
    loads the required internal and external modules then subsequently performs
    apischema deserialization to produce a list of component configuration objects.

    Args:
        config_path ([type]): The path to the config file.

    Returns:
        List[Component]: A list of component configuration objects.
    """
    yaml_struct = yaml.load(open(config_path, "r"), Loader=yaml.Loader)
    configs = deserialize(
        List[ComponentConfig],
        yaml_struct,
        default_conversion=importing_conversion,
    )
    return configs



from typing import Any, DefaultDict, Dict, Iterator, Type, TypeVar

from apischema import deserializer
from apischema.conversions import Conversion
from apischema.tagged_unions import Tagged, TaggedUnion, get_tagged

# Implementation adapted from apischema example: Class as tagged union of its subclasses
# see: https://wyfo.github.io/apischema/examples/subclass_tagged_union/

#: A class
Cls = TypeVar("Cls", bound=type)


def rec_subclasses(cls: type) -> Iterator[type]:
    """Recursive implementation of type.__subclasses__.

    Args:
        cls (Type): The base class.

    Returns:
        Iterator[type]: An iterator of subclasses.
    """
    for sub_cls in cls.__subclasses__():
        yield sub_cls
        yield from rec_subclasses(sub_cls)


#: Whether the current class is registered as a tagged union
is_tagged_union: Dict[Type[Any], bool] = DefaultDict(lambda: False)


def as_tagged_union(cls: Cls) -> Cls:
    """A decorator to make a config base class which can deserialize aliased
    sub-classes.

    A decorator which makes a config class the root of a tagged union of sub-classes
    allowing for serialization and deserialization of config trees by class alias. The
    function registers both an apischema serialization and an apischema deserialization
    conversion for the base class which perform lookup based on a tagged union of
    aliased sub-classes.

    Args:
        cls (Cls): The config base class.

    Returns:
        Cls: The modified config base class.
    """

    # This will only be used if we want to generate a json schema (which we will)
    def deserialization() -> Conversion:
        annotations: Dict[str, Any] = {}
        deserialization_namespace: Dict[str, Any] = {"__annotations__": annotations}
        for sub in rec_subclasses(cls):
            fullname = sub.__module__ + "." + sub.__name__
            annotations[fullname] = Tagged[sub]  # type: ignore
        deserialization_union = type(
            cls.__name__ + "TaggedUnion",
            (TaggedUnion,),
            deserialization_namespace,
        )
        return Conversion(
            lambda obj: get_tagged(obj)[1], source=deserialization_union, target=cls
        )

    deserializer(lazy=deserialization, target=cls)
    is_tagged_union[cls] = True
    return cls



import logging
from typing import Iterable

import pytest
from mock import Mock, patch

from tickit.core.typedefs import SimTime
from tickit.devices.sink import SinkDevice


@pytest.fixture
def sink() -> SinkDevice:
    return SinkDevice()


@pytest.fixture
def mock_logging() -> Iterable[Mock]:
    with patch("tickit.devices.sink.LOGGER", autospec=True) as mock:
        yield mock


def test_sink_update_method(sink: SinkDevice, caplog):
    inputs = SinkDevice.Inputs({"input": "blah"})
    with caplog.at_level(logging.DEBUG):
        device_update = sink.update(SimTime(0), inputs)
    assert device_update.outputs == {}
    assert device_update.call_at is None

    assert len(caplog.records) == 1
    record: logging.LogRecord = caplog.records[0]

    assert record.levelname == "DEBUG"
    assert record.message == "Sunk {}".format(inputs)



import re
from typing import AnyStr, AsyncIterable, List, Tuple

from tickit.adapters.interpreters.utils import wrap_messages_as_async_iterable
from tickit.core.adapter import Adapter, Interpreter


class SplittingInterpreter(Interpreter[AnyStr]):
    """A wrapper for an interpreter that splits a single message into multiple.

    An interpreter wrapper class that takes a message, splits it according to a given
    delimiter, and passes on the resulting sub-messages individually on to the
    wrapped interpreter.
    """

    def __init__(
        self,
        interpreter: Interpreter[AnyStr],
        message_delimiter: AnyStr,
    ) -> None:
        """A decorator for an interpreter that splits a message into multiple
        sub-messages.

        Args:
            interpreter (Interpreter): The interpreter messages are passed on to.
            message_delimiter (AnyStr): The delimiter by which the message is split up.
                Can be a regex pattern. Must be of the same type as the message.
        """
        super().__init__()
        self.interpreter: Interpreter[AnyStr] = interpreter
        self.message_delimiter: AnyStr = message_delimiter

    async def _handle_individual_messages(
        self, adapter: Adapter, individual_messages: List[AnyStr]
    ) -> List[Tuple[AsyncIterable[AnyStr], bool]]:
        results = [
            await self.interpreter.handle(adapter, message)
            for message in individual_messages
        ]
        return results

    async def _collect_responses(
        self, results: List[Tuple[AsyncIterable[AnyStr], bool]]
    ) -> Tuple[AsyncIterable[AnyStr], bool]:
        """Combines results from handling multiple messages.

        Takes the responses from when the wrapped interpreter handles multiple messages
        and returns an appropriate composite repsonse and interrrupt. The response is
        an asyncronous iterable of each of the individual responses, the composite
        interrupt is a logical inclusive 'or' of all of the individual interrupts.

        Args:
            results (List[Tuple[AsyncIterable[AnyStr], bool]]): a list of returned
                values from the wrapped class' handle() method.

        Returns:
            Tuple[AsyncIterable[AnyStr], bool]:
                A tuple of the asynchronous iterable of reply messages and a flag
                indicating whether an interrupt should be raised by the adapter.
        """
        individual_responses, individual_interrupts = zip(*results)

        responses = [
            response
            for response_gen in individual_responses
            async for response in response_gen
        ]

        resp = wrap_messages_as_async_iterable(responses)
        interrrupt = any(individual_interrupts)
        return resp, interrrupt

    async def handle(
        self, adapter: Adapter, message: AnyStr
    ) -> Tuple[AsyncIterable[AnyStr], bool]:
        """Splits a message and passes the resulting sub-messages on to an interpreter.

        Splits a given message and passes the resulting sub-messages on to an
        interpreter. The responses to the individual sub-messages are then returned.

        Args:
            adapter (Adapter): The adapter in which the function should be executed
            message: (AnyStr): The message to be split up and handled.

        Returns:
            Tuple[AsyncIterable[Union[str, bytes]], bool]:
                A tuple of the asynchronous iterable of reply messages and a flag
                indicating whether an interrupt should be raised by the adapter.
        """
        individual_messages = [
            msg for msg in re.split(self.message_delimiter, message) if msg is not None
        ]

        results = await self._handle_individual_messages(adapter, individual_messages)

        return await self._collect_responses(results)



from tickit.adapters.interpreters.wrappers.beheading_interpreter import (
    BeheadingInterpreter,
)
from tickit.adapters.interpreters.wrappers.joining_interpreter import JoiningInterpreter
from tickit.adapters.interpreters.wrappers.splitting_interpreter import (
    SplittingInterpreter,
)

__all__ = ["BeheadingInterpreter", "SplittingInterpreter", "JoiningInterpreter"]



from dataclasses import dataclass, field
from enum import Enum
from functools import partial
from typing import Any, Generic, List, Mapping, Optional, TypeVar

T = TypeVar("T")


def field_config(**kwargs) -> Mapping[str, Any]:
    """Helper function to create a typesafe dictionary.

    Helper function to create a typesafe dictionary to be inserted as
    dataclass metadata.

    Args:
        kwargs: Key/value pairs to go into the metadata

    Returns:
        Mapping[str, Any]: A dictionary of {key: value} where all keys are strings
    """
    return dict(**kwargs)


class AccessMode(Enum):
    """Possible access modes for field metadata."""

    READ_ONLY: str = "r"
    WRITE_ONLY: str = "w"
    READ_WRITE: str = "rw"
    FLOAT: str = "float"
    INT: str = "int"
    UINT: str = "uint"
    STRING: str = "string"
    LIST_STR: str = "string[]"
    BOOL: str = "bool"
    FLOAT_GRID: str = "float[][]"
    UINT_GRID: str = "uint[][]"
    DATE: str = "date"
    NONE: str = "none"


#
# Shortcuts to creating dataclass field metadata
#
rw_float: partial = partial(
    field_config, value_type=AccessMode.FLOAT, access_mode=AccessMode.READ_WRITE
)
ro_float: partial = partial(
    field_config, value_type=AccessMode.FLOAT, access_mode=AccessMode.READ_ONLY
)
rw_int: partial = partial(
    field_config, value_type=AccessMode.INT, access_mode=AccessMode.READ_WRITE
)
ro_int: partial = partial(
    field_config, value_type=AccessMode.INT, access_mode=AccessMode.READ_ONLY
)
rw_uint: partial = partial(
    field_config, value_type=AccessMode.UINT, access_mode=AccessMode.READ_WRITE
)
rw_str: partial = partial(
    field_config, value_type=AccessMode.STRING, access_mode=AccessMode.READ_WRITE
)
ro_str: partial = partial(
    field_config, value_type=AccessMode.STRING, access_mode=AccessMode.READ_ONLY
)
rw_bool: partial = partial(
    field_config, value_type=AccessMode.BOOL, access_mode=AccessMode.READ_WRITE
)
rw_float_grid: partial = partial(
    field_config,
    value_type=AccessMode.FLOAT_GRID,
    access_mode=AccessMode.READ_WRITE,
)
rw_uint_grid: partial = partial(
    field_config,
    value_type=AccessMode.UINT_GRID,
    access_mode=AccessMode.READ_WRITE,
)
ro_date: partial = partial(
    field_config, value_type=AccessMode.DATE, access_mode=AccessMode.READ_ONLY
)


@dataclass
class Value(Generic[T]):
    """Schema for a value to be returned by the API. Most fields are optional."""

    value: T
    value_type: str
    access_mode: Optional[AccessMode] = None
    unit: Optional[str] = None
    min: Optional[T] = None
    max: Optional[T] = None
    allowed_values: Optional[List[str]] = None


@dataclass
class SequenceComplete:
    """Schema for confirmation returned by operations that do not return values."""

    sequence_id: int = field(default=1, metadata=ro_int())



import sys
from typing import Awaitable, Callable, Dict, Iterable, Set, Tuple, Type, TypeVar
from warnings import warn

# TODO: Investigate why import from tickit.utils.compat.typing_compat causes mypy error
# See mypy issue for details: https://github.com/python/mypy/issues/10851
if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
elif sys.version_info >= (3, 5):
    from typing_extensions import Protocol, runtime_checkable

#: A consumable value
C = TypeVar("C", covariant=True)
#: A producable value
P = TypeVar("P", contravariant=True)


@runtime_checkable
class StateConsumer(Protocol[C]):
    """An interface for types which implent publish/subscribe message consumers.

    An interface for types which implement publish/subscribe message consumers,
    the consumer must be able to subscribe to topics within the messaging framework,
    upon recieving a message the consumer should pass the value to the callback
    function, if a topic is subscribed to which does not yet exist it should be created.
    """

    def __init__(self, callback: Callable[[C], Awaitable[None]]) -> None:
        """Constructor of the consumer, given a callback handle.

        Args:
            callback (Callable[[C], Awaitable[None]]): An asynchronous handler function
                for consumed values.
        """
        pass

    async def subscribe(self, topics: Iterable[str]) -> None:
        """Subscribe the consumer to the given topics.

        Subscribe the consumer to the given topics, new messages are passed to
        the callback.

        Args:
            topics (Iterable[str]): An iterable of topics to subscribe to.
        """
        pass


@runtime_checkable
class StateProducer(Protocol[P]):
    """An interface for types which implement publish/subscribe message producers.

    An interface for types which implment publish/subscribe message producers,
    the producer must be able to produce a value to a topic within the messaging
    framework, if the topic does not yet exist it should be created.
    """

    def __init__(self) -> None:
        """Constructor of the producer, given no arguments."""
        pass

    async def produce(self, topic: str, value: P) -> None:
        """Produce a value to the provided topic.

        Args:
            topic (str): The topic to which the value should be sent.
            value (P): The value to send to the provided topic.
        """
        pass


#: The union of StateConsumer and StateProducer
StateInterface = TypeVar("StateInterface", StateConsumer, StateProducer)


consumers: Dict[str, Tuple[Type[StateConsumer], bool]] = dict()
producers: Dict[str, Tuple[Type[StateProducer], bool]] = dict()


def add(
    name: str, external: bool
) -> Callable[[Type[StateInterface]], Type[StateInterface]]:
    """A decorator to add a StateInterface to the registry.

    A decorator to add a StateInterface to the registry of StateConsumers or
    StateProducer according to it's signature. StateConsumers and StateProducers which
    are intended to work together should be added with the same name.

    Args:
        name (str): The name under which the class should be registered (typically the
            name of the messaging framework).
        external (bool): A flag which indicates whether the interface can be used
            simulations which are distributed across processes.
    """

    def wrap(interface: Type[StateInterface]) -> Type[StateInterface]:
        if isinstance(interface, StateProducer):
            producers[name] = (interface, external)
        elif isinstance(interface, StateConsumer):
            consumers[name] = (interface, external)
        else:
            warn(
                RuntimeWarning(
                    "{} is not {} or {}".format(interface, StateConsumer, StateProducer)
                )
            )
        return interface

    return wrap


def interfaces(external: bool = False) -> Set[str]:
    """Returns interface names for which both a StateConsumer and StateProducer exist.

    Gets a set of interface names for which both a StateConsumer and StateProducer
    exist. The external option may be used to restrict interfaces to those which may be
    used in simulations which are distriubted across multiple processes.

    Args:
        external (bool): If true, only interfaces which can be used in simulations
            which are distributed across processes are returned. If false, all
            interfaces are returned. Defaults to False.

    Returns:
        Set[str]: A set of names of StateConsumer / StateProducer pairs.
    """
    return satisfy_externality(external, consumers) & satisfy_externality(
        external, producers
    )


def satisfy_externality(
    external: bool, interfaces: Dict[str, Tuple[Type[StateInterface], bool]]
) -> Set[str]:
    """Finds which interfaces satisfy the externality requirement provided.

    Args:
        external (bool): If true, only interfaces which can be used in
            simulations which are distributed across processes are returned. If false,
            all interfaces are returned.
        interfaces (Dict[str, Tuple[Type[StateInterface], bool]]): A mapping of
            interface names to a tuple of their class and their externality flag.

    Returns:
        Set[str]: A set of interface names which satisfy the externality requirement.
    """
    return set(
        name for name, interface in interfaces.items() if not external or interface[1]
    )


def get_interface(name: str) -> Tuple[Type[StateConsumer], Type[StateProducer]]:
    """Get the StateConsumer and StateProducer classes for a given interface name.

    Args:
        name (str): The name of the interface to be retrieved.

    Returns:
        Tuple[Type[StateConsumer], Type[StateProducer]]:
            A tuple of the StateConsumer and StateProducer classes.
    """
    return consumers[name][0], producers[name][0]



from types import new_class
from typing import Type

from _pytest.fixtures import fixture

from tickit.utils.singleton import Singleton


@fixture
def TestSingleton() -> Type:
    return new_class("TestSingleton", tuple(), {"metaclass": Singleton})


@fixture
def OtherSingleton() -> Type:
    return new_class("OtherSingleton", tuple(), {"metaclass": Singleton})


def test_call_creates(TestSingleton):
    assert isinstance(TestSingleton(), object)


def subsequent_call_creates_same(TestSingleton):
    original = TestSingleton()
    assert original == TestSingleton()


def derived_unequal(TestSingleton, OtherSigleton):
    assert TestSingleton() != OtherSigleton()



import pytest

from tickit.utils.byte_format import ByteFormat


@pytest.fixture
def byte_format() -> ByteFormat:
    return ByteFormat(b"%b/r/n")


def test_byte_format_constructor(byte_format: ByteFormat):
    pass


def test_byteformat_serializer(byte_format: ByteFormat):
    serialized = byte_format.serialize()
    assert "%b/r/n" == serialized


def test_byteformat_deserializer(byte_format: ByteFormat):
    deserialized = byte_format.deserialize("%b/r/n")
    assert ByteFormat(b"%b/r/n") == deserialized



import asyncio
import logging

import aiozmq
import pytest
from mock import Mock
from mock.mock import AsyncMock, create_autospec

from tickit.adapters.zmqadapter import ZeroMQAdapter
from tickit.core.device import Device


@pytest.fixture
def mock_device() -> Device:
    return create_autospec(Device)


@pytest.fixture
def mock_raise_interrupt():
    async def raise_interrupt():
        return False

    return Mock(raise_interrupt)


@pytest.fixture
def mock_process_message_queue() -> AsyncMock:
    async def _process_message_queue():
        return True

    return AsyncMock(_process_message_queue)


@pytest.fixture
def zeromq_adapter() -> ZeroMQAdapter:
    zmq_adapter = ZeroMQAdapter()
    zmq_adapter._dealer = AsyncMock()
    zmq_adapter._router = AsyncMock()
    zmq_adapter._message_queue = Mock(asyncio.Queue)
    return zmq_adapter


def test_zeromq_adapter_constructor():
    ZeroMQAdapter()


@pytest.mark.asyncio
async def test_zeromq_adapter_start_stream(zeromq_adapter: ZeroMQAdapter):
    await zeromq_adapter.start_stream()

    assert isinstance(zeromq_adapter._router, aiozmq.stream.ZmqStream)
    assert isinstance(zeromq_adapter._dealer, aiozmq.stream.ZmqStream)

    await zeromq_adapter.close_stream()


@pytest.mark.asyncio
async def test_zeromq_adapter_close_stream(zeromq_adapter: ZeroMQAdapter):
    await zeromq_adapter.start_stream()

    await zeromq_adapter.close_stream()
    await asyncio.sleep(0.1)

    assert None is zeromq_adapter._router._transport
    assert None is zeromq_adapter._dealer._transport


@pytest.mark.asyncio
async def test_zeromq_adapter_after_update(zeromq_adapter):
    zeromq_adapter.after_update()


@pytest.mark.asyncio
async def test_zeromq_adapter_send_message(zeromq_adapter):
    mock_message = AsyncMock()

    zeromq_adapter.send_message(mock_message)
    task = asyncio.current_task()
    asyncio.gather(task)
    zeromq_adapter._message_queue.put.assert_called_once()


@pytest.mark.asyncio
async def test_zeromq_adapter_run_forever_method(
    zeromq_adapter,
    mock_device: Device,
    mock_process_message_queue: AsyncMock,
    mock_raise_interrupt: Mock,
):
    zeromq_adapter._process_message_queue = mock_process_message_queue

    await zeromq_adapter.run_forever(mock_device, mock_raise_interrupt)

    zeromq_adapter._process_message_queue.assert_called_once()


@pytest.mark.asyncio
async def test_zeromq_adapter_check_if_running(zeromq_adapter):
    assert zeromq_adapter.check_if_running() is False


@pytest.mark.asyncio
async def test_zeromq_adapter_process_message_queue(zeromq_adapter):
    zeromq_adapter._process_message = AsyncMock()
    zeromq_adapter.check_if_running = Mock(return_value=False)

    await zeromq_adapter._process_message_queue()

    zeromq_adapter._process_message.assert_awaited_once()


@pytest.mark.asyncio
async def test_zeromq_adapter_process_message(zeromq_adapter):
    mock_message = "test"

    zeromq_adapter._dealer.read.return_value = ("Data", "test")
    zeromq_adapter._router.read.return_value = ("Data", "test")

    await zeromq_adapter._process_message(mock_message)

    zeromq_adapter._dealer.read.assert_awaited_once()
    zeromq_adapter._router.read.assert_awaited_once()


@pytest.mark.asyncio
async def test_zeromq_adapter_process_message_no_message(zeromq_adapter, caplog):
    mock_message = None

    zeromq_adapter._dealer.read.return_value = ("Data", None)
    zeromq_adapter._router.read.return_value = ("Data", None)

    with caplog.at_level(logging.DEBUG):
        await zeromq_adapter._process_message(mock_message)

    assert len(caplog.records) == 1



from tickit.core.state_interfaces import internal, kafka
from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer

__all__ = ["StateConsumer", "StateProducer", "internal", "kafka"]



import asyncio
import traceback
from typing import Awaitable, Iterable, List


def run_all(awaitables: Iterable[Awaitable[None]]) -> List[asyncio.Task]:
    """Asynchronously runs the run_forever method of each lifetime runnable.

    Creates and runs an asyncio task for the run_forever method of each lifetime
    runnable. Calls to the run_forever method are wrapped with an error handler.

    Args:
        awaitables: An iterable of awaitable objects.

    Returns:
        List[asyncio.Task]: A list of asyncio tasks for the awaitables.
    """

    async def run_with_error_handling(awaitable: Awaitable[None]) -> None:
        try:
            await awaitable
        except Exception as e:
            # These are prints rather than logging because we just want the
            # result going directly to stdout.
            print("Task exception: {}".format(e))
            print(traceback.format_exc())

    return [
        asyncio.create_task(run_with_error_handling(awaitable))
        for awaitable in awaitables
    ]


async def run_all_forever(awaitables: Iterable[Awaitable[None]]) -> None:
    """Asynchronously runs the run_forever method of each lifetime runnable.

    Creates and runs an asyncio task for the run_forever method of each lifetime
    runnable. Calls to the run_forever method are wrapped with an error handler.
    This function blocks until all run_forever methods have completed.

    Args:
        awaitables: An iterable of awaitable objects.
    """
    await asyncio.wait(run_all(awaitables))



from dataclasses import dataclass, field, fields
from typing import Any

from tickit.devices.eiger.eiger_schema import rw_int, rw_str


@dataclass
class MonitorConfig:
    """Eiger monitor configuration taken from the API spec."""

    mode: str = field(
        default="enabled", metadata=rw_str(allowed_values=["disabled", "enabled"])
    )
    buffer_size: int = field(default=512, metadata=rw_int())

    def __getitem__(self, key: str) -> Any:  # noqa: D105
        f = {}
        for field_ in fields(self):
            f[field_.name] = {
                "value": vars(self)[field_.name],
                "metadata": field_.metadata,
            }
        return f[key]



from typing import Any, Iterable

import pytest
import pytest_asyncio
from immutables import Map
from mock import Mock, create_autospec, patch

from tickit.core.management.event_router import Wiring
from tickit.core.management.schedulers.base import BaseScheduler
from tickit.core.state_interfaces.state_interface import StateConsumer, StateProducer
from tickit.core.typedefs import (
    Changes,
    ComponentID,
    Input,
    Interrupt,
    Output,
    PortID,
    SimTime,
)


@pytest.fixture
def patch_base_schedule_interrupt_method() -> Iterable[Mock]:
    with patch.object(BaseScheduler, "schedule_interrupt") as mock:
        yield mock


@pytest.fixture
def mock_wiring() -> Mock:
    return create_autospec(Wiring, instance=True)


@pytest.fixture
def mock_state_consumer_type() -> Mock:
    return create_autospec(StateConsumer, instance=False)


@pytest.fixture
def mock_state_producer_type() -> Mock:
    return create_autospec(StateProducer, instance=False)


@pytest.fixture
def patch_ticker() -> Iterable[Mock]:
    with patch("tickit.core.management.schedulers.base.Ticker", autospec=True) as mock:
        yield mock


@pytest_asyncio.fixture
async def base_scheduler(
    mock_wiring: Mock,
    mock_state_consumer_type,
    mock_state_producer_type,
    patch_ticker,
    patch_base_schedule_interrupt_method,
) -> BaseScheduler:
    base_scheduler = BaseScheduler(  # type: ignore
        mock_wiring, mock_state_consumer_type, mock_state_producer_type
    )
    await base_scheduler.setup()
    return base_scheduler


def test_base_scheduler_constructor_and_setup(base_scheduler):
    pass


@pytest.mark.asyncio
async def test_base_scheduler_update_component_method(
    base_scheduler: Any,
):
    _input = Input(
        ComponentID("foo"), SimTime(0), Changes(Map({PortID("42"): "hello"}))
    )
    await base_scheduler.update_component(_input)
    base_scheduler.state_producer.produce.assert_awaited_once()


@pytest.mark.asyncio
async def test_base_scheduler_handle_output_message(base_scheduler: Any):
    message = Output(
        ComponentID("bar"),
        SimTime(42),
        Changes(Map({PortID("42"): "world"})),
        SimTime(88),
    )
    await base_scheduler.handle_message(message)
    base_scheduler.ticker.propagate.assert_awaited_once_with(message)


@pytest.mark.asyncio
async def test_base_scheduler_handle_interrupt_message(base_scheduler: BaseScheduler):
    message = Interrupt(ComponentID("foo"))
    await base_scheduler.handle_message(message)
    base_scheduler.schedule_interrupt.assert_called_once()  # type: ignore


def test_base_scheduler_get_first_wakeups_method(base_scheduler: BaseScheduler):
    expected_component = ComponentID("foo")
    expected_when = SimTime(42)
    base_scheduler.add_wakeup(expected_component, expected_when)

    components, when = base_scheduler.get_first_wakeups()

    assert components == {expected_component}
    assert when == expected_when


def test_base_scheduler_get_empty_wakeups(base_scheduler: BaseScheduler):
    components, when = base_scheduler.get_first_wakeups()
    assert components == set()
    assert when is None
